import numpy as np
from .common import Benchmark, safe_import

with safe_import():
    from scipy import stats
with safe_import():
    from scipy.stats import sampling
with safe_import():
    from scipy import special


# Beta distribution with a = 2, b = 3
class contdist1:
    def __init__(self):
        self.mode = 1/3

    def pdf(self, x):
        return 12 * x * (1-x)**2

    def dpdf(self, x):
        return 12 * ((1-x)**2 - 2*x*(1-x))

    def cdf(self, x):
        return 12 * (x**2/2 - x**3/3 + x**4/4)

    def support(self):
        return 0, 1

    def __repr__(self):
        # asv prints this.
        return 'beta(2, 3)'


# Standard Normal Distribution
class contdist2:
    def __init__(self):
        self.mode = 0

    def pdf(self, x):
        return 1./np.sqrt(2*np.pi) * np.exp(-0.5 * x*x)

    def dpdf(self, x):
        return 1./np.sqrt(2*np.pi) * -x * np.exp(-0.5 * x*x)

    def cdf(self, x):
        return special.ndtr(x)

    def __repr__(self):
        return 'norm(0, 1)'


# pdf with piecewise linear function as transformed density with T = -1/sqrt
# Taken from UNU.RAN test suite (from file t_tdr_ps.c)
class contdist3:
    def __init__(self, shift=0.):
        self.shift = shift
        self.mode = shift

    def pdf(self, x):
        x -= self.shift
        y = 1. / (abs(x) + 1.)
        return y * y

    def dpdf(self, x):
        x -= self.shift
        y = 1. / (abs(x) + 1.)
        y = 2. * y * y * y
        return y if (x < 0.) else -y

    def cdf(self, x):
        x -= self.shift
        if x <= 0.:
            return 0.5 / (1. - x)
        return 1. - 0.5 / (1. + x)

    def __repr__(self):
        return f'sqrtlinshft({self.shift})'


# Sin 2 distribution
#          /  0.05 + 0.45*(1 +sin(2 Pi x))  if |x| <= 1
#  f(x) = <
#          \  0        otherwise
# Taken from UNU.RAN test suite (from file t_pinv.c)
class contdist4:
    def __init__(self):
        self.mode = 0

    def pdf(self, x):
        return 0.05 + 0.45 * (1 + np.sin(2*np.pi*x))

    def dpdf(self, x):
        return 0.2 * 0.45 * (2*np.pi) * np.cos(2*np.pi*x)

    def cdf(self, x):
        return (0.05*(x + 1) +
                0.9*(1. + 2.*np.pi*(1 + x) - np.cos(2.*np.pi*x)) /
                (4.*np.pi))

    def support(self):
        return -1, 1

    def __repr__(self):
        return 'sin2'


# Sin 10 distribution
#          /  0.05 + 0.45*(1 +sin(2 Pi x))  if |x| <= 5
#  f(x) = <
#          \  0        otherwise
# Taken from UNU.RAN test suite (from file t_pinv.c)
class contdist5:
    def __init__(self):
        self.mode = 0

    def pdf(self, x):
        return 0.2 * (0.05 + 0.45 * (1 + np.sin(2*np.pi*x)))

    def dpdf(self, x):
        return 0.2 * 0.45 * (2*np.pi) * np.cos(2*np.pi*x)

    def cdf(self, x):
        return x/10. + 0.5 + 0.09/(2*np.pi) * (np.cos(10*np.pi) -
                                               np.cos(2*np.pi*x))

    def support(self):
        return -5, 5

    def __repr__(self):
        return 'sin10'


allcontdists = [contdist1(), contdist2(), contdist3(), contdist3(10000.),
                contdist4(), contdist5()]


class TransformedDensityRejection(Benchmark):

    param_names = ['dist', 'c']

    params = [allcontdists, [0., -0.5]]

    def setup(self, dist, c):
        self.urng = np.random.default_rng(0xfaad7df1c89e050200dbe258636b3265)
        with np.testing.suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            try:
                self.rng = sampling.TransformedDensityRejection(
                    dist, c=c, random_state=self.urng
                )
            except sampling.UNURANError:
                # contdist3 is not T-concave for c=0. So, skip such test-cases
                raise NotImplementedError(f"{dist} not T-concave for c={c}")

    def time_tdr_setup(self, dist, c):
        with np.testing.suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            sampling.TransformedDensityRejection(
                dist, c=c, random_state=self.urng
            )

    def time_tdr_rvs(self, dist, c):
        self.rng.rvs(100000)


class SimpleRatioUniforms(Benchmark):

    param_names = ['dist', 'cdf_at_mode']

    params = [allcontdists, [0, 1]]

    def setup(self, dist, cdf_at_mode):
        self.urng = np.random.default_rng(0xfaad7df1c89e050200dbe258636b3265)
        try:
            if cdf_at_mode:
                cdf_at_mode = dist.cdf(dist.mode)
            else:
                cdf_at_mode = None
            self.rng = sampling.SimpleRatioUniforms(
                dist, mode=dist.mode,
                cdf_at_mode=cdf_at_mode,
                random_state=self.urng
            )
        except sampling.UNURANError:
            raise NotImplementedError(f"{dist} not T-concave")

    def time_srou_setup(self, dist, cdf_at_mode):
        if cdf_at_mode:
            cdf_at_mode = dist.cdf(dist.mode)
        else:
            cdf_at_mode = None
        sampling.SimpleRatioUniforms(
            dist, mode=dist.mode,
            cdf_at_mode=cdf_at_mode,
            random_state=self.urng
        )

    def time_srou_rvs(self, dist, cdf_at_mode):
        self.rng.rvs(100000)


class NumericalInversePolynomial(Benchmark):

    param_names = ['dist']

    params = [allcontdists]

    def setup(self, dist):
        self.urng = np.random.default_rng(0xb235b58c1f616c59c18d8568f77d44d1)
        with np.testing.suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            try:
                self.rng = sampling.NumericalInversePolynomial(
                    dist, random_state=self.urng
                )
            except sampling.UNURANError:
                raise NotImplementedError(f"setup failed for {dist}")

    def time_pinv_setup(self, dist):
        with np.testing.suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            sampling.NumericalInversePolynomial(
                dist, random_state=self.urng
            )

    def time_pinv_rvs(self, dist):
        self.rng.rvs(100000)


class NumericalInverseHermite(Benchmark):

    param_names = ['dist', 'order']
    params = [allcontdists, [3, 5]]

    def setup(self, dist, order):
        self.urng = np.random.default_rng(0xb235b58c1f616c59c18d8568f77d44d1)
        with np.testing.suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            try:
                self.rng = sampling.NumericalInverseHermite(
                    dist, order=order, random_state=self.urng
                )
            except sampling.UNURANError:
                raise NotImplementedError(f"setup failed for {dist}")

    def time_hinv_setup(self, dist, order):
        with np.testing.suppress_warnings() as sup:
            sup.filter(RuntimeWarning)
            sampling.NumericalInverseHermite(
                dist, order=order, random_state=self.urng
            )

    def time_hinv_rvs(self, dist, order):
        self.rng.rvs(100000)


class DiscreteAliasUrn(Benchmark):

    param_names = ['distribution']

    params = [
        # a subset of discrete distributions with finite domain.
        [['nhypergeom', (20, 7, 1)],
         ['hypergeom', (30, 12, 6)],
         ['nchypergeom_wallenius', (140, 80, 60, 0.5)],
         ['binom', (5, 0.4)]]
    ]

    def setup(self, distribution):
        distname, params = distribution
        dist = getattr(stats, distname)
        domain = dist.support(*params)
        self.urng = np.random.default_rng(0x2fc9eb71cd5120352fa31b7a048aa867)
        x = np.arange(domain[0], domain[1] + 1)
        self.pv = dist.pmf(x, *params)
        self.rng = sampling.DiscreteAliasUrn(self.pv, random_state=self.urng)

    def time_dau_setup(self, distribution):
        sampling.DiscreteAliasUrn(self.pv, random_state=self.urng)

    def time_dau_rvs(self, distribution):
        self.rng.rvs(100000)


class DiscreteGuideTable(Benchmark):

    param_names = ['distribution']

    params = [
        # a subset of discrete distributions with finite domain.
        [['nhypergeom', (20, 7, 1)],
         ['hypergeom', (30, 12, 6)],
         ['nchypergeom_wallenius', (140, 80, 60, 0.5)],
         ['binom', (5, 0.4)]]
    ]

    def setup(self, distribution):
        distname, params = distribution
        dist = getattr(stats, distname)
        domain = dist.support(*params)
        self.urng = np.random.default_rng(0x2fc9eb71cd5120352fa31b7a048aa867)
        x = np.arange(domain[0], domain[1] + 1)
        self.pv = dist.pmf(x, *params)
        self.rng = sampling.DiscreteGuideTable(self.pv, random_state=self.urng)

    def time_dgt_setup(self, distribution):
        sampling.DiscreteGuideTable(self.pv, random_state=self.urng)

    def time_dgt_rvs(self, distribution):
        self.rng.rvs(100000)



"""Functions to extract parts of sparse matrices
"""

__docformat__ = "restructuredtext en"

__all__ = ['find', 'tril', 'triu']


from ._coo import coo_matrix, coo_array
from ._base import sparray


def find(A):
    """Return the indices and values of the nonzero elements of a matrix

    Parameters
    ----------
    A : dense or sparse array or matrix
        Matrix whose nonzero elements are desired.

    Returns
    -------
    (I,J,V) : tuple of arrays
        I,J, and V contain the row indices, column indices, and values
        of the nonzero entries.


    Examples
    --------
    >>> from scipy.sparse import csr_array, find
    >>> A = csr_array([[7.0, 8.0, 0],[0, 0, 9.0]])
    >>> find(A)
    (array([0, 0, 1], dtype=int32),
     array([0, 1, 2], dtype=int32),
     array([ 7.,  8.,  9.]))

    """

    A = coo_array(A, copy=True)
    A.sum_duplicates()
    # remove explicit zeros
    nz_mask = A.data != 0
    return A.row[nz_mask], A.col[nz_mask], A.data[nz_mask]


def tril(A, k=0, format=None):
    """Return the lower triangular portion of a sparse array or matrix

    Returns the elements on or below the k-th diagonal of A.
        - k = 0 corresponds to the main diagonal
        - k > 0 is above the main diagonal
        - k < 0 is below the main diagonal

    Parameters
    ----------
    A : dense or sparse array or matrix
        Matrix whose lower trianglar portion is desired.
    k : integer : optional
        The top-most diagonal of the lower triangle.
    format : string
        Sparse format of the result, e.g. format="csr", etc.

    Returns
    -------
    L : sparse matrix
        Lower triangular portion of A in sparse format.

    See Also
    --------
    triu : upper triangle in sparse format

    Examples
    --------
    >>> from scipy.sparse import csr_array, tril
    >>> A = csr_array([[1, 2, 0, 0, 3], [4, 5, 0, 6, 7], [0, 0, 8, 9, 0]],
    ...               dtype='int32')
    >>> A.toarray()
    array([[1, 2, 0, 0, 3],
           [4, 5, 0, 6, 7],
           [0, 0, 8, 9, 0]], dtype=int32)
    >>> tril(A).toarray()
    array([[1, 0, 0, 0, 0],
           [4, 5, 0, 0, 0],
           [0, 0, 8, 0, 0]], dtype=int32)
    >>> tril(A).nnz
    4
    >>> tril(A, k=1).toarray()
    array([[1, 2, 0, 0, 0],
           [4, 5, 0, 0, 0],
           [0, 0, 8, 9, 0]], dtype=int32)
    >>> tril(A, k=-1).toarray()
    array([[0, 0, 0, 0, 0],
           [4, 0, 0, 0, 0],
           [0, 0, 0, 0, 0]], dtype=int32)
    >>> tril(A, format='csc')
    <Compressed Sparse Column sparse array of dtype 'int32'
        with 4 stored elements and shape (3, 5)>

    """
    coo_sparse = coo_array if isinstance(A, sparray) else coo_matrix

    # convert to COOrdinate format where things are easy
    A = coo_sparse(A, copy=False)
    mask = A.row + k >= A.col

    row = A.row[mask]
    col = A.col[mask]
    data = A.data[mask]
    new_coo = coo_sparse((data, (row, col)), shape=A.shape, dtype=A.dtype)
    return new_coo.asformat(format)


def triu(A, k=0, format=None):
    """Return the upper triangular portion of a sparse array or matrix

    Returns the elements on or above the k-th diagonal of A.
        - k = 0 corresponds to the main diagonal
        - k > 0 is above the main diagonal
        - k < 0 is below the main diagonal

    Parameters
    ----------
    A : dense or sparse array or matrix
        Matrix whose upper trianglar portion is desired.
    k : integer : optional
        The bottom-most diagonal of the upper triangle.
    format : string
        Sparse format of the result, e.g. format="csr", etc.

    Returns
    -------
    L : sparse array or matrix
        Upper triangular portion of A in sparse format.
        Sparse array if A is a sparse array, otherwise matrix.

    See Also
    --------
    tril : lower triangle in sparse format

    Examples
    --------
    >>> from scipy.sparse import csr_array, triu
    >>> A = csr_array([[1, 2, 0, 0, 3], [4, 5, 0, 6, 7], [0, 0, 8, 9, 0]],
    ...                dtype='int32')
    >>> A.toarray()
    array([[1, 2, 0, 0, 3],
           [4, 5, 0, 6, 7],
           [0, 0, 8, 9, 0]], dtype=int32)
    >>> triu(A).toarray()
    array([[1, 2, 0, 0, 3],
           [0, 5, 0, 6, 7],
           [0, 0, 8, 9, 0]], dtype=int32)
    >>> triu(A).nnz
    8
    >>> triu(A, k=1).toarray()
    array([[0, 2, 0, 0, 3],
           [0, 0, 0, 6, 7],
           [0, 0, 0, 9, 0]], dtype=int32)
    >>> triu(A, k=-1).toarray()
    array([[1, 2, 0, 0, 3],
           [4, 5, 0, 6, 7],
           [0, 0, 8, 9, 0]], dtype=int32)
    >>> triu(A, format='csc')
    <Compressed Sparse Column sparse array of dtype 'int32'
        with 8 stored elements and shape (3, 5)>

    """
    coo_sparse = coo_array if isinstance(A, sparray) else coo_matrix

    # convert to COOrdinate format where things are easy
    A = coo_sparse(A, copy=False)
    mask = A.row + k <= A.col

    row = A.row[mask]
    col = A.col[mask]
    data = A.data[mask]
    new_coo = coo_sparse((data, (row, col)), shape=A.shape, dtype=A.dtype)
    return new_coo.asformat(format)



"""
Dogleg algorithm with rectangular trust regions for least-squares minimization.

The description of the algorithm can be found in [Voglis]_. The algorithm does
trust-region iterations, but the shape of trust regions is rectangular as
opposed to conventional elliptical. The intersection of a trust region and
an initial feasible region is again some rectangle. Thus, on each iteration a
bound-constrained quadratic optimization problem is solved.

A quadratic problem is solved by well-known dogleg approach, where the
function is minimized along piecewise-linear "dogleg" path [NumOpt]_,
Chapter 4. If Jacobian is not rank-deficient then the function is decreasing
along this path, and optimization amounts to simply following along this
path as long as a point stays within the bounds. A constrained Cauchy step
(along the anti-gradient) is considered for safety in rank deficient cases,
in this situations the convergence might be slow.

If during iterations some variable hit the initial bound and the component
of anti-gradient points outside the feasible region, then a next dogleg step
won't make any progress. At this state such variables satisfy first-order
optimality conditions and they are excluded before computing a next dogleg
step.

Gauss-Newton step can be computed exactly by `numpy.linalg.lstsq` (for dense
Jacobian matrices) or by iterative procedure `scipy.sparse.linalg.lsmr` (for
dense and sparse matrices, or Jacobian being LinearOperator). The second
option allows to solve very large problems (up to couple of millions of
residuals on a regular PC), provided the Jacobian matrix is sufficiently
sparse. But note that dogbox is not very good for solving problems with
large number of constraints, because of variables exclusion-inclusion on each
iteration (a required number of function evaluations might be high or accuracy
of a solution will be poor), thus its large-scale usage is probably limited
to unconstrained problems.

References
----------
.. [Voglis] C. Voglis and I. E. Lagaris, "A Rectangular Trust Region Dogleg
            Approach for Unconstrained and Bound Constrained Nonlinear
            Optimization", WSEAS International Conference on Applied
            Mathematics, Corfu, Greece, 2004.
.. [NumOpt] J. Nocedal and S. J. Wright, "Numerical optimization, 2nd edition".
"""
import numpy as np
from numpy.linalg import lstsq, norm

from scipy.sparse.linalg import LinearOperator, aslinearoperator, lsmr
from scipy.optimize import OptimizeResult
from scipy._lib._util import _call_callback_maybe_halt


from .common import (
    step_size_to_bound, in_bounds, update_tr_radius, evaluate_quadratic,
    build_quadratic_1d, minimize_quadratic_1d, compute_grad,
    compute_jac_scale, check_termination, scale_for_robust_loss_function,
    print_header_nonlinear, print_iteration_nonlinear)


def lsmr_operator(Jop, d, active_set):
    """Compute LinearOperator to use in LSMR by dogbox algorithm.

    `active_set` mask is used to excluded active variables from computations
    of matrix-vector products.
    """
    m, n = Jop.shape

    def matvec(x):
        x_free = x.ravel().copy()
        x_free[active_set] = 0
        return Jop.matvec(x * d)

    def rmatvec(x):
        r = d * Jop.rmatvec(x)
        r[active_set] = 0
        return r

    return LinearOperator((m, n), matvec=matvec, rmatvec=rmatvec, dtype=float)


def find_intersection(x, tr_bounds, lb, ub):
    """Find intersection of trust-region bounds and initial bounds.

    Returns
    -------
    lb_total, ub_total : ndarray with shape of x
        Lower and upper bounds of the intersection region.
    orig_l, orig_u : ndarray of bool with shape of x
        True means that an original bound is taken as a corresponding bound
        in the intersection region.
    tr_l, tr_u : ndarray of bool with shape of x
        True means that a trust-region bound is taken as a corresponding bound
        in the intersection region.
    """
    lb_centered = lb - x
    ub_centered = ub - x

    lb_total = np.maximum(lb_centered, -tr_bounds)
    ub_total = np.minimum(ub_centered, tr_bounds)

    orig_l = np.equal(lb_total, lb_centered)
    orig_u = np.equal(ub_total, ub_centered)

    tr_l = np.equal(lb_total, -tr_bounds)
    tr_u = np.equal(ub_total, tr_bounds)

    return lb_total, ub_total, orig_l, orig_u, tr_l, tr_u


def dogleg_step(x, newton_step, g, a, b, tr_bounds, lb, ub):
    """Find dogleg step in a rectangular region.

    Returns
    -------
    step : ndarray, shape (n,)
        Computed dogleg step.
    bound_hits : ndarray of int, shape (n,)
        Each component shows whether a corresponding variable hits the
        initial bound after the step is taken:
            *  0 - a variable doesn't hit the bound.
            * -1 - lower bound is hit.
            *  1 - upper bound is hit.
    tr_hit : bool
        Whether the step hit the boundary of the trust-region.
    """
    lb_total, ub_total, orig_l, orig_u, tr_l, tr_u = find_intersection(
        x, tr_bounds, lb, ub
    )
    bound_hits = np.zeros_like(x, dtype=int)

    if in_bounds(newton_step, lb_total, ub_total):
        return newton_step, bound_hits, False

    to_bounds, _ = step_size_to_bound(np.zeros_like(x), -g, lb_total, ub_total)

    # The classical dogleg algorithm would check if Cauchy step fits into
    # the bounds, and just return it constrained version if not. But in a
    # rectangular trust region it makes sense to try to improve constrained
    # Cauchy step too. Thus, we don't distinguish these two cases.

    cauchy_step = -minimize_quadratic_1d(a, b, 0, to_bounds)[0] * g

    step_diff = newton_step - cauchy_step
    step_size, hits = step_size_to_bound(cauchy_step, step_diff,
                                         lb_total, ub_total)
    bound_hits[(hits < 0) & orig_l] = -1
    bound_hits[(hits > 0) & orig_u] = 1
    tr_hit = np.any((hits < 0) & tr_l | (hits > 0) & tr_u)

    return cauchy_step + step_size * step_diff, bound_hits, tr_hit


def dogbox(fun, jac, x0, f0, J0, lb, ub, ftol, xtol, gtol, max_nfev, x_scale,
           loss_function, tr_solver, tr_options, verbose, callback=None):
    f = f0
    f_true = f.copy()
    nfev = 1

    J = J0
    njev = 1

    if loss_function is not None:
        rho = loss_function(f)
        cost = 0.5 * np.sum(rho[0])
        J, f = scale_for_robust_loss_function(J, f, rho)
    else:
        cost = 0.5 * np.dot(f, f)

    g = compute_grad(J, f)

    jac_scale = isinstance(x_scale, str) and x_scale == 'jac'
    if jac_scale:
        scale, scale_inv = compute_jac_scale(J)
    else:
        scale, scale_inv = x_scale, 1 / x_scale

    Delta = norm(x0 * scale_inv, ord=np.inf)
    if Delta == 0:
        Delta = 1.0

    on_bound = np.zeros_like(x0, dtype=int)
    on_bound[np.equal(x0, lb)] = -1
    on_bound[np.equal(x0, ub)] = 1

    x = x0
    step = np.empty_like(x0)

    if max_nfev is None:
        max_nfev = x0.size * 100

    termination_status = None
    iteration = 0
    step_norm = None
    actual_reduction = None

    if verbose == 2:
        print_header_nonlinear()

    while True:
        active_set = on_bound * g < 0
        free_set = ~active_set

        g_free = g[free_set]
        g_full = g.copy()
        g[active_set] = 0

        g_norm = norm(g, ord=np.inf)
        if g_norm < gtol:
            termination_status = 1

        if verbose == 2:
            print_iteration_nonlinear(iteration, nfev, cost, actual_reduction,
                                      step_norm, g_norm)

        if termination_status is not None or nfev == max_nfev:
            break

        x_free = x[free_set]
        lb_free = lb[free_set]
        ub_free = ub[free_set]
        scale_free = scale[free_set]

        # Compute (Gauss-)Newton and build quadratic model for Cauchy step.
        if tr_solver == 'exact':
            J_free = J[:, free_set]
            newton_step = lstsq(J_free, -f, rcond=-1)[0]

            # Coefficients for the quadratic model along the anti-gradient.
            a, b = build_quadratic_1d(J_free, g_free, -g_free)
        elif tr_solver == 'lsmr':
            Jop = aslinearoperator(J)

            # We compute lsmr step in scaled variables and then
            # transform back to normal variables, if lsmr would give exact lsq
            # solution, this would be equivalent to not doing any
            # transformations, but from experience it's better this way.

            # We pass active_set to make computations as if we selected
            # the free subset of J columns, but without actually doing any
            # slicing, which is expensive for sparse matrices and impossible
            # for LinearOperator.

            lsmr_op = lsmr_operator(Jop, scale, active_set)
            newton_step = -lsmr(lsmr_op, f, **tr_options)[0][free_set]
            newton_step *= scale_free

            # Components of g for active variables were zeroed, so this call
            # is correct and equivalent to using J_free and g_free.
            a, b = build_quadratic_1d(Jop, g, -g)

        actual_reduction = -1.0
        while actual_reduction <= 0 and nfev < max_nfev:
            tr_bounds = Delta * scale_free

            step_free, on_bound_free, tr_hit = dogleg_step(
                x_free, newton_step, g_free, a, b, tr_bounds, lb_free, ub_free)

            step.fill(0.0)
            step[free_set] = step_free

            if tr_solver == 'exact':
                predicted_reduction = -evaluate_quadratic(J_free, g_free,
                                                          step_free)
            elif tr_solver == 'lsmr':
                predicted_reduction = -evaluate_quadratic(Jop, g, step)

            # gh11403 ensure that solution is fully within bounds.
            x_new = np.clip(x + step, lb, ub)

            f_new = fun(x_new)
            nfev += 1

            step_h_norm = norm(step * scale_inv, ord=np.inf)

            if not np.all(np.isfinite(f_new)):
                Delta = 0.25 * step_h_norm
                continue

            # Usual trust-region step quality estimation.
            if loss_function is not None:
                cost_new = loss_function(f_new, cost_only=True)
            else:
                cost_new = 0.5 * np.dot(f_new, f_new)
            actual_reduction = cost - cost_new

            Delta, ratio = update_tr_radius(
                Delta, actual_reduction, predicted_reduction,
                step_h_norm, tr_hit
            )

            step_norm = norm(step)
            termination_status = check_termination(
                actual_reduction, cost, step_norm, norm(x), ratio, ftol, xtol)

            if termination_status is not None:
                break

        if actual_reduction > 0:
            on_bound[free_set] = on_bound_free

            x = x_new
            # Set variables exactly at the boundary.
            mask = on_bound == -1
            x[mask] = lb[mask]
            mask = on_bound == 1
            x[mask] = ub[mask]

            f = f_new
            f_true = f.copy()

            cost = cost_new

            J = jac(x)
            njev += 1

            if loss_function is not None:
                rho = loss_function(f)
                J, f = scale_for_robust_loss_function(J, f, rho)

            g = compute_grad(J, f)

            if jac_scale:
                scale, scale_inv = compute_jac_scale(J, scale_inv)
        else:
            step_norm = 0
            actual_reduction = 0

        iteration += 1
        
        # Call callback function and possibly stop optimization
        if callback is not None:
            intermediate_result = OptimizeResult(
                x=x, fun=f, nit=iteration, nfev=nfev)
            intermediate_result["cost"] = cost_new

            if _call_callback_maybe_halt(
                callback, intermediate_result
            ):
                termination_status = -2
                break

    if termination_status is None:
        termination_status = 0

    return OptimizeResult(
        x=x, cost=cost, fun=f_true, jac=J, grad=g_full, optimality=g_norm,
        active_mask=on_bound, nfev=nfev, njev=njev, status=termination_status)



#!/usr/bin/env python3
"""
Minimize the Rosenbrock function subject to simple bounds and randomly
generated linear inequality and equality constraints.
"""
import numpy as np
from cobyqa import minimize
from scipy.optimize import Bounds, LinearConstraint, rosen


if __name__ == "__main__":
    rng = np.random.default_rng(0)
    n, m_linear_ub, m_linear_eq = 10, 3, 2

    # Generate an initial guess satisfying the bound constraints.
    bounds = Bounds(-2.048 * np.ones(n), 2.048 * np.ones(n))
    x0 = rng.uniform(-3.0, 3.0, n)
    x0 = np.maximum(bounds.lb, np.minimum(bounds.ub, x0))

    # Generate feasible linear inequality and equality constraints.
    x_rand = rng.uniform(bounds.lb, bounds.ub)
    aub = rng.standard_normal((m_linear_ub, n))
    bub = np.dot(aub, x_rand) + rng.uniform(0.0, 1.0, m_linear_ub)
    aeq = rng.standard_normal((m_linear_eq, n))
    beq = np.dot(aeq, x_rand)
    constraints = [
        LinearConstraint(aub, -np.inf, bub),
        LinearConstraint(aeq, beq, beq),
    ]
    res = minimize(rosen, x0, bounds=bounds, constraints=constraints)
    print(res)



import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
import numpy as np

from scipy import signal
from scipy.fft import rfft, rfftfreq

# Create input signal:
n = 50
x = np.zeros(n)
x[0] = n

# Apply FIR filter which delays signal by 3 samples:
y = signal.lfilter([0, 0, 0, 1], 1, x)

X, Y = (rfft(z_) / n for z_ in (x, y))
f = rfftfreq(n, 1)  # sampling interval T = 1 s

fig = plt.figure(tight_layout=True, figsize=(6., 4.))
gs = gridspec.GridSpec(3, 1)
ax0 = fig.add_subplot(gs[0, :])
ax1 = fig.add_subplot(gs[1:, :], sharex=ax0)

for Z_, n_, m_ in zip((X, Y), ("Input $X(f)$", "Output $Y(f)$"), ('+-', 'x-')):
    ax0.plot(f, abs(Z_), m_, alpha=.5, label=n_)
    ax1.plot(f, np.unwrap(np.angle(Z_)), m_, alpha=.5, label=n_)

ax0.set(title="Frequency Response of 3 Sample Delay Filter (no window)",
        ylabel="Magnitude", xlim=(0, f[-1]), ylim=(0, 1.1))
ax1.set(xlabel=rf"Frequency $f$ in Hertz ($\Delta f = {1/n}\,$Hz)",
        ylabel=r"Phase in rad")
ax1.set_yticks(-np.arange(0, 7)*np.pi/2,
               ['0', '-π/2', '-π', '-3/2π', '-2π', '-4/3π', '-3π'])

ax2 = ax1.twinx()
ax2.set(ylabel=r"Phase in Degrees", ylim=np.rad2deg(ax1.get_ylim()),
        yticks=np.arange(-540, 90, 90))
for ax_ in (ax0, ax1):
    ax_.legend()
    ax_.grid()
plt.show()



"""
Sparse Eigenvalue Solvers
-------------------------

The submodules of sparse.linalg._eigen:
    1. lobpcg: Locally Optimal Block Preconditioned Conjugate Gradient Method

"""
from .arpack import *
from .lobpcg import *
from ._svds import svds

from . import arpack

__all__ = [
    'ArpackError', 'ArpackNoConvergence',
    'eigs', 'eigsh', 'lobpcg', 'svds'
]

from scipy._lib._testutils import PytestTester
test = PytestTester(__name__)
del PytestTester



'''
This is an example to illustrate the usage of the solver.

Translated from Zaikun Zhang's modern-Fortran reference implementation in PRIMA.

Dedicated to late Professor M. J. D. Powell FRS (1936--2015).

Python translation by Nickolai Belakovski.
'''

import numpy as np
from pyprima.cobyla.cobyla import cobyla


def test_chebyquad():
    f, _ = calcfc_chebyquad([1, 2])
    assert np.isclose(f, 91+1/9, atol=1e-6)


def calcfc_chebyquad(x):
    x = np.array(x)
    n = len(x)  # or shape?
    y = np.zeros((n + 1, n + 1))
    y[:n, 0] = 1
    y[:n, 1] = 2*x - 1
    for i in range(1, n):
        y[:n, i+1] = 2*y[:n, 1] * y[:n, i] - y[:n, i - 1]

    f = 0
    for i in range(n+1):
        tmp = sum(y[0:n, i]) / n
        if i % 2 == 0:
            tmp += 1/((i)**2 - 1)
        f += tmp**2
    constr = np.zeros(0)
    return f, constr

def calcfc_hexagon(x):
    # Test problem 10 in Powell's original algorithm

    assert len(x) == 9

    f = -0.5 * (x[0] * x[3] - x[1] * x[2] + x[2] * x[8] - x[4] * x[8] + x[4] * x[7] - x[5] * x[6])
    constr = np.zeros(14)
    constr[0] = -1 + x[2]**2 + x[3]**2
    constr[1] = -1 + x[8]**2
    constr[2] = -1 + x[4]**2 + x[5]**2
    constr[3] = -1 + x[0]**2 + (x[1] - x[8])**2
    constr[4] = -1 + (x[0] - x[4])**2 + (x[1] - x[5])**2
    constr[5] = -1 + (x[0] - x[6])**2 + (x[1] - x[7])**2
    constr[6] = -1 + (x[2] - x[4])**2 + (x[3] - x[5])**2
    constr[7] = -1 + (x[2] - x[6])**2 + (x[3] - x[7])**2
    constr[8] = -1 + x[6]**2 + (x[7] - x[8])**2
    constr[9] = -x[0] * x[3] + x[1] * x[2]
    constr[10] = -x[2] * x[8]
    constr[11] = -x[4] * x[8]
    constr[12] = -x[4] * x[7] + x[5] * x[6]
    constr[13] = -x[8]

    return f, constr


if __name__ == "__main__":
    n_chebyquad = 6

    # The following lines illustrates how to call the solver to solve the Chebyquad problem.
    x_chebyquad = np.array([i/(n_chebyquad+1) for i in range(1, n_chebyquad+1)])  # Starting point
    m = 0  # Dimension of constraints. M must be specified correctly, or the program will crash!
    result = cobyla(calcfc_chebyquad, m, x_chebyquad)  # This call will not print anything

    # In addition to the compulsory arguments, the following illustration specifies also CONSTR, RHOBEG,
    # and IPRINT, which are optional. All the unspecified optional arguments (RHOEND, MAXFUN, etc.) will
    # take their default values coded in the solver.
    x_chebyquad = np.array([i/(n_chebyquad+1) for i in range(1, n_chebyquad+1)])  # Starting point
    result = cobyla(calcfc_chebyquad, m, x_chebyquad, rhobeg=0.2 * x_chebyquad[0], iprint=1)

    # The following lines illustrates how to call the solver to solve the Hexagon problem.
    x_hexagon = np.zeros(9) + 2
    m_hexagon = 14  # Dimension of constraints. M must the specified correctly, or the program will crash!
    result = cobyla(calcfc_hexagon, m_hexagon, x_hexagon)  # This call will not print anything.

    # In addition to the compulsory arguments, the following illustration specifies also CONSTR, RHOBEG,
    # and IPRINT, which are optional. All the unspecified optional arguments (RHOEND, MAXFUN, etc.) will
    # take their default values coded in the solver. Note that CONSTR is an output, which will be set to
    # the value of CONSTR(X_HEXAGON) when the solver returns.
    result = cobyla(calcfc_hexagon, m_hexagon, x_hexagon, rhobeg=1.0, iprint=1)



import pytest
from array_api_compat import device, to_device

xp = pytest.importorskip("array_api_compat.cupy")
from cupy.cuda import Stream


@pytest.mark.parametrize(
    "make_stream",
    [
        lambda: Stream(),
        lambda: Stream(non_blocking=True),
        lambda: Stream(null=True),
        lambda: Stream(ptds=True),
    ],
)
def test_to_device_with_stream(make_stream):
    devices = xp.__array_namespace_info__().devices()

    a = xp.asarray([1, 2, 3])
    for dev in devices:
        # Streams are device-specific and must be created within
        # the context of the device...
        with dev:
            stream = make_stream()
        # ... however, to_device() does not need to be inside the
        # device context.
        b = to_device(a, dev, stream=stream)
        assert device(b) == dev


def test_to_device_with_dlpack_stream():
    devices = xp.__array_namespace_info__().devices()

    a = xp.asarray([1, 2, 3])
    for dev in devices:
        # Streams are device-specific and must be created within
        # the context of the device...
        with dev:
            s1 = Stream()

        # ... however, to_device() does not need to be inside the
        # device context.
        b = to_device(a, dev, stream=s1.ptr)
        assert device(b) == dev



"""Trust-region interior point method.

References
----------
.. [1] Byrd, Richard H., Mary E. Hribar, and Jorge Nocedal.
       "An interior point algorithm for large-scale nonlinear
       programming." SIAM Journal on Optimization 9.4 (1999): 877-900.
.. [2] Byrd, Richard H., Guanghui Liu, and Jorge Nocedal.
       "On the local behavior of an interior point method for
       nonlinear programming." Numerical analysis 1997 (1997): 37-56.
.. [3] Nocedal, Jorge, and Stephen J. Wright. "Numerical optimization"
       Second Edition (2006).
"""

import scipy.sparse as sps
import numpy as np
from .equality_constrained_sqp import equality_constrained_sqp
from scipy.sparse.linalg import LinearOperator

__all__ = ['tr_interior_point']


class BarrierSubproblem:
    """
    Barrier optimization problem:
        minimize fun(x) - barrier_parameter*sum(log(s))
        subject to: constr_eq(x)     = 0
                  constr_ineq(x) + s = 0
    """

    def __init__(self, x0, s0, fun, grad, lagr_hess, n_vars, n_ineq, n_eq,
                 constr, jac, barrier_parameter, tolerance,
                 enforce_feasibility, global_stop_criteria,
                 xtol, fun0, grad0, constr_ineq0, jac_ineq0, constr_eq0,
                 jac_eq0, finite_diff_bounds):
        # Store parameters
        self.n_vars = n_vars
        self.x0 = x0
        self.s0 = s0
        self.fun = fun
        self.grad = grad
        self.lagr_hess = lagr_hess
        self.constr = constr
        self.jac = jac
        self.barrier_parameter = barrier_parameter
        self.tolerance = tolerance
        self.n_eq = n_eq
        self.n_ineq = n_ineq
        self.enforce_feasibility = enforce_feasibility
        self.global_stop_criteria = global_stop_criteria
        self.xtol = xtol
        self.fun0 = self._compute_function(fun0, constr_ineq0, s0)
        self.grad0 = self._compute_gradient(grad0)
        self.constr0 = self._compute_constr(constr_ineq0, constr_eq0, s0)
        self.jac0 = self._compute_jacobian(jac_eq0, jac_ineq0, s0)
        self.terminate = False
        self.lb = finite_diff_bounds[0]
        self.ub = finite_diff_bounds[1]

    def update(self, barrier_parameter, tolerance):
        self.barrier_parameter = barrier_parameter
        self.tolerance = tolerance

    def get_slack(self, z):
        return z[self.n_vars:self.n_vars+self.n_ineq]

    def get_variables(self, z):
        return z[:self.n_vars]

    def function_and_constraints(self, z):
        """Returns barrier function and constraints at given point.

        For z = [x, s], returns barrier function:
            function(z) = fun(x) - barrier_parameter*sum(log(s))
        and barrier constraints:
            constraints(z) = [   constr_eq(x)     ]
                             [ constr_ineq(x) + s ]

        """
        # Get variables and slack variables
        x = self.get_variables(z)
        s = self.get_slack(z)

        # Compute function and constraints,
        # making sure x is within any strict bounds
        if np.any((x < self.lb) | (x > self.ub)):
            # If x is out of the strict bounds, set f = inf,
            # and just set both equality and inequality
            # constraints to 0 since we can't evaluate
            # them separately.
            f = np.inf
            c_eq = np.full(self.n_eq, 0.)
            c_ineq = np.full(self.n_ineq, 0.)
        else:
            f = self.fun(x)
            c_eq, c_ineq = self.constr(x)

        # Return objective function and constraints
        return (self._compute_function(f, c_ineq, s),
                self._compute_constr(c_ineq, c_eq, s))

    def _compute_function(self, f, c_ineq, s):
        # Use technique from Nocedal and Wright book, ref [3]_, p.576,
        # to guarantee constraints from `enforce_feasibility`
        # stay feasible along iterations.
        s[self.enforce_feasibility] = -c_ineq[self.enforce_feasibility]
        log_s = [np.log(s_i) if s_i > 0 else -np.inf for s_i in s]
        # Compute barrier objective function
        return f - self.barrier_parameter*np.sum(log_s)

    def _compute_constr(self, c_ineq, c_eq, s):
        # Compute barrier constraint
        return np.hstack((c_eq,
                          c_ineq + s))

    def scaling(self, z):
        """Returns scaling vector.
        Given by:
            scaling = [ones(n_vars), s]
        """
        s = self.get_slack(z)
        diag_elements = np.hstack((np.ones(self.n_vars), s))

        # Diagonal matrix
        def matvec(vec):
            return diag_elements*vec
        return LinearOperator((self.n_vars+self.n_ineq,
                               self.n_vars+self.n_ineq),
                              matvec)

    def gradient_and_jacobian(self, z):
        """Returns scaled gradient.

        Return scaled gradient:
            gradient = [             grad(x)             ]
                       [ -barrier_parameter*ones(n_ineq) ]
        and scaled Jacobian matrix:
            jacobian = [  jac_eq(x)  0  ]
                       [ jac_ineq(x) S  ]
        Both of them scaled by the previously defined scaling factor.
        """
        # Get variables and slack variables
        x = self.get_variables(z)
        s = self.get_slack(z)
        # Compute first derivatives
        g = self.grad(x)
        J_eq, J_ineq = self.jac(x)
        # Return gradient and Jacobian
        return (self._compute_gradient(g),
                self._compute_jacobian(J_eq, J_ineq, s))

    def _compute_gradient(self, g):
        return np.hstack((g, -self.barrier_parameter*np.ones(self.n_ineq)))

    def _compute_jacobian(self, J_eq, J_ineq, s):
        if self.n_ineq == 0:
            return J_eq
        else:
            if sps.issparse(J_eq) or sps.issparse(J_ineq):
                # It is expected that J_eq and J_ineq
                # are already `csr_array` because of
                # the way ``BoxConstraint``, ``NonlinearConstraint``
                # and ``LinearConstraint`` are defined.
                J_eq = sps.csr_array(J_eq)
                J_ineq = sps.csr_array(J_ineq)
                return self._assemble_sparse_jacobian(J_eq, J_ineq, s)
            else:
                S = np.diag(s)
                zeros = np.zeros((self.n_eq, self.n_ineq))
                # Convert to matrix
                if sps.issparse(J_ineq):
                    J_ineq = J_ineq.toarray()
                if sps.issparse(J_eq):
                    J_eq = J_eq.toarray()
                # Concatenate matrices
                return np.block([[J_eq, zeros],
                                 [J_ineq, S]])

    def _assemble_sparse_jacobian(self, J_eq, J_ineq, s):
        """Assemble sparse Jacobian given its components.

        Given ``J_eq``, ``J_ineq`` and ``s`` returns:
            jacobian = [ J_eq,     0     ]
                       [ J_ineq, diag(s) ]

        It is equivalent to:
            sps.bmat([[ J_eq,   None    ],
                      [ J_ineq, diag(s) ]], "csr")
        but significantly more efficient for this
        given structure.
        """
        n_vars, n_ineq, n_eq = self.n_vars, self.n_ineq, self.n_eq
        J_aux = sps.vstack([J_eq, J_ineq], "csr")
        indptr, indices, data = J_aux.indptr, J_aux.indices, J_aux.data
        new_indptr = indptr + np.hstack((np.zeros(n_eq, dtype=int),
                                         np.arange(n_ineq+1, dtype=int)))
        size = indices.size+n_ineq
        new_indices = np.empty(size)
        new_data = np.empty(size)
        mask = np.full(size, False, bool)
        mask[new_indptr[-n_ineq:]-1] = True
        new_indices[mask] = n_vars+np.arange(n_ineq)
        new_indices[~mask] = indices
        new_data[mask] = s
        new_data[~mask] = data
        J = sps.csr_array((new_data, new_indices, new_indptr),
                          (n_eq + n_ineq, n_vars + n_ineq))
        return J

    def lagrangian_hessian_x(self, z, v):
        """Returns Lagrangian Hessian (in relation to `x`) -> Hx"""
        x = self.get_variables(z)
        # Get lagrange multipliers related to nonlinear equality constraints
        v_eq = v[:self.n_eq]
        # Get lagrange multipliers related to nonlinear ineq. constraints
        v_ineq = v[self.n_eq:self.n_eq+self.n_ineq]
        lagr_hess = self.lagr_hess
        return lagr_hess(x, v_eq, v_ineq)

    def lagrangian_hessian_s(self, z, v):
        """Returns scaled Lagrangian Hessian (in relation to`s`) -> S Hs S"""
        s = self.get_slack(z)
        # Using the primal formulation:
        #     S Hs S = diag(s)*diag(barrier_parameter/s**2)*diag(s).
        # Reference [1]_ p. 882, formula (3.1)
        primal = self.barrier_parameter
        # Using the primal-dual formulation
        #     S Hs S = diag(s)*diag(v/s)*diag(s)
        # Reference [1]_ p. 883, formula (3.11)
        primal_dual = v[-self.n_ineq:]*s
        # Uses the primal-dual formulation for
        # positives values of v_ineq, and primal
        # formulation for the remaining ones.
        return np.where(v[-self.n_ineq:] > 0, primal_dual, primal)

    def lagrangian_hessian(self, z, v):
        """Returns scaled Lagrangian Hessian"""
        # Compute Hessian in relation to x and s
        Hx = self.lagrangian_hessian_x(z, v)
        if self.n_ineq > 0:
            S_Hs_S = self.lagrangian_hessian_s(z, v)

        # The scaled Lagragian Hessian is:
        #     [ Hx    0    ]
        #     [ 0   S Hs S ]
        def matvec(vec):
            vec_x = self.get_variables(vec)
            vec_s = self.get_slack(vec)
            if self.n_ineq > 0:
                return np.hstack((Hx.dot(vec_x), S_Hs_S*vec_s))
            else:
                return Hx.dot(vec_x)
        return LinearOperator((self.n_vars+self.n_ineq,
                               self.n_vars+self.n_ineq),
                              matvec)

    def stop_criteria(self, state, z, last_iteration_failed,
                      optimality, constr_violation,
                      trust_radius, penalty, cg_info):
        """Stop criteria to the barrier problem.
        The criteria here proposed is similar to formula (2.3)
        from [1]_, p.879.
        """
        x = self.get_variables(z)
        if self.global_stop_criteria(state, x,
                                     last_iteration_failed,
                                     trust_radius, penalty,
                                     cg_info,
                                     self.barrier_parameter,
                                     self.tolerance):
            self.terminate = True
            return True
        else:
            g_cond = (optimality < self.tolerance and
                      constr_violation < self.tolerance)
            x_cond = trust_radius < self.xtol
            return g_cond or x_cond


def tr_interior_point(fun, grad, lagr_hess, n_vars, n_ineq, n_eq,
                      constr, jac, x0, fun0, grad0,
                      constr_ineq0, jac_ineq0, constr_eq0,
                      jac_eq0, stop_criteria,
                      enforce_feasibility, xtol, state,
                      initial_barrier_parameter,
                      initial_tolerance,
                      initial_penalty,
                      initial_trust_radius,
                      factorization_method,
                      finite_diff_bounds):
    """Trust-region interior points method.

    Solve problem:
        minimize fun(x)
        subject to: constr_ineq(x) <= 0
                    constr_eq(x) = 0
    using trust-region interior point method described in [1]_.
    """
    # BOUNDARY_PARAMETER controls the decrease on the slack
    # variables. Represents ``tau`` from [1]_ p.885, formula (3.18).
    BOUNDARY_PARAMETER = 0.995
    # BARRIER_DECAY_RATIO controls the decay of the barrier parameter
    # and of the subproblem tolerance. Represents ``theta`` from [1]_ p.879.
    BARRIER_DECAY_RATIO = 0.2
    # TRUST_ENLARGEMENT controls the enlargement on trust radius
    # after each iteration
    TRUST_ENLARGEMENT = 5

    # Default enforce_feasibility
    if enforce_feasibility is None:
        enforce_feasibility = np.zeros(n_ineq, bool)
    # Initial Values
    barrier_parameter = initial_barrier_parameter
    tolerance = initial_tolerance
    trust_radius = initial_trust_radius
    # Define initial value for the slack variables
    s0 = np.maximum(-1.5*constr_ineq0, np.ones(n_ineq))
    # Define barrier subproblem
    subprob = BarrierSubproblem(
        x0, s0, fun, grad, lagr_hess, n_vars, n_ineq, n_eq, constr, jac,
        barrier_parameter, tolerance, enforce_feasibility,
        stop_criteria, xtol, fun0, grad0, constr_ineq0, jac_ineq0,
        constr_eq0, jac_eq0, finite_diff_bounds)
    # Define initial parameter for the first iteration.
    z = np.hstack((x0, s0))
    fun0_subprob, constr0_subprob = subprob.fun0, subprob.constr0
    grad0_subprob, jac0_subprob = subprob.grad0, subprob.jac0
    # Define trust region bounds
    trust_lb = np.hstack((np.full(subprob.n_vars, -np.inf),
                          np.full(subprob.n_ineq, -BOUNDARY_PARAMETER)))
    trust_ub = np.full(subprob.n_vars+subprob.n_ineq, np.inf)

    # Solves a sequence of barrier problems
    while True:
        # Solve SQP subproblem
        z, state = equality_constrained_sqp(
            subprob.function_and_constraints,
            subprob.gradient_and_jacobian,
            subprob.lagrangian_hessian,
            z, fun0_subprob, grad0_subprob,
            constr0_subprob, jac0_subprob, subprob.stop_criteria,
            state, initial_penalty, trust_radius,
            factorization_method, trust_lb, trust_ub, subprob.scaling)
        if subprob.terminate:
            break
        # Update parameters
        trust_radius = max(initial_trust_radius,
                           TRUST_ENLARGEMENT*state.tr_radius)
        # TODO: Use more advanced strategies from [2]_
        # to update this parameters.
        barrier_parameter *= BARRIER_DECAY_RATIO
        tolerance *= BARRIER_DECAY_RATIO
        # Update Barrier Problem
        subprob.update(barrier_parameter, tolerance)
        # Compute initial values for next iteration
        fun0_subprob, constr0_subprob = subprob.function_and_constraints(z)
        grad0_subprob, jac0_subprob = subprob.gradient_and_jacobian(z)

    # Get x and s
    x = subprob.get_variables(z)
    return x, state



#!/usr/bin/env python3
r"""
Solve the feasibility problem

.. math::

    (y + x^2)^2 + 0.1 y^2 \le 1,\\
    y \le e^{-x} - 3,\\
    y \le x - 4.

"""
import numpy as np
from cobyqa import minimize
from scipy.optimize import LinearConstraint, NonlinearConstraint


def cub(x):
    return [(x[1] + x[0] ** 2) ** 2 + 0.1 * x[1] ** 2, x[1] - np.exp(-x[0])]


if __name__ == "__main__":
    x0 = [0.0, 0.0]
    constraints = [
        LinearConstraint([[-1.0, 1.0]], -np.inf, [-4.0]),
        NonlinearConstraint(cub, -np.inf, [1.0, -3.0]),
    ]
    options = {
        "disp": True,
        "feasibility_tol": np.finfo(float).eps,
    }
    minimize(None, x0, constraints=constraints, options=options)



""" Benchmark linalg.sqrtm for various blocksizes.

"""
import numpy as np

from .common import Benchmark, safe_import

with safe_import():
    import scipy.linalg


class Sqrtm(Benchmark):
    params = [
        ['float64', 'complex128'],
        [16, 32, 64, 256, 512],
    ]
    param_names = ['dtype', 'n']

    def setup(self, dtype, n):
        n = int(n)
        rng = np.random.default_rng(1742808411247533)
        dtype = np.dtype(dtype)
        A = rng.uniform(size=[n, n])
        if dtype == np.complex128:
            A = A + 1j*rng.uniform(size=[n, n])
        self.A = A

    def time_sqrtm(self, dtype, n):
        scipy.linalg.sqrtm(self.A, disp=False)



# cython: boundscheck=False, wraparound=False, cdivision=True
import numpy as np


ctypedef struct Pair:
    int key
    double value


cdef class Heap:
    """Binary heap.

    Heap stores values and keys. Values are passed explicitly, whereas keys
    are assigned implicitly to natural numbers (from 0 to n - 1).

    The supported operations (all have O(log n) time complexity):

        * Return the current minimum value and the corresponding key.
        * Remove the current minimum value.
        * Change the value of the given key. Note that the key must be still
          in the heap.

    The heap is stored as an array, where children of parent i have indices
    2 * i + 1 and 2 * i + 2. All public methods are based on  `sift_down` and
    `sift_up` methods, which restore the heap property by moving an element
    down or up in the heap.
    """
    cdef int[:] index_by_key
    cdef int[:] key_by_index
    cdef double[:] values
    cdef int size

    def __init__(self, double[:] values):
        self.size = values.shape[0]
        self.index_by_key = np.arange(self.size, dtype=np.intc)
        self.key_by_index = np.arange(self.size, dtype=np.intc)
        self.values = values.copy()
        cdef int i

        # Create the heap in a linear time. The algorithm sequentially sifts
        # down items starting from lower levels.
        for i in reversed(range(self.size / 2)):
            self.sift_down(i)

    cpdef Pair get_min(self) noexcept:
        return Pair(self.key_by_index[0], self.values[0])

    cpdef void remove_min(self) noexcept:
        self.swap(0, self.size - 1)
        self.size -= 1
        self.sift_down(0)

    cpdef void change_value(self, int key, double value) noexcept:
        cdef int index = self.index_by_key[key]
        cdef double old_value = self.values[index]
        self.values[index] = value
        if value < old_value:
            self.sift_up(index)
        else:
            self.sift_down(index)

    cdef void sift_up(self, int index) noexcept:
        cdef int parent = Heap.parent(index)
        while index > 0 and self.values[parent] > self.values[index]:
            self.swap(index, parent)
            index = parent
            parent = Heap.parent(index)

    cdef void sift_down(self, int index) noexcept:
        cdef int child = Heap.left_child(index)
        while child < self.size:
            if (child + 1 < self.size and
                    self.values[child + 1] < self.values[child]):
                child += 1

            if self.values[index] > self.values[child]:
                self.swap(index, child)
                index = child
                child = Heap.left_child(index)
            else:
                break

    @staticmethod
    cdef inline int left_child(int parent) noexcept:
        return (parent << 1) + 1

    @staticmethod
    cdef inline int parent(int child) noexcept:
        return (child - 1) >> 1

    cdef void swap(self, int i, int j) noexcept:
        self.values[i], self.values[j] = self.values[j], self.values[i]
        cdef int key_i = self.key_by_index[i]
        cdef int key_j = self.key_by_index[j]
        self.key_by_index[i] = key_j
        self.key_by_index[j] = key_i
        self.index_by_key[key_i] = j
        self.index_by_key[key_j] = i



""" Benchmark linalg.logm for various blocksizes.

"""
import numpy as np
from .common import Benchmark, safe_import

with safe_import():
    import scipy.linalg


class Logm(Benchmark):
    params = [
        ['float64', 'complex128'],
        [64, 256],
        ['gen', 'her', 'pos']
    ]
    param_names = ['dtype', 'n', 'structure']

    def setup(self, dtype, n, structure):
        n = int(n)
        dtype = np.dtype(dtype)

        A = np.random.rand(n, n)
        if dtype == np.complex128:
            A = A + 1j*np.random.rand(n, n)

        if structure == 'pos':
            A = A @ A.T.conj()
        elif structure == 'her':
            A = A + A.T.conj()

        self.A = A

    def time_logm(self, dtype, n, structure):
        scipy.linalg.logm(self.A, disp=False)



from typing import Final

from dask.array import *  # noqa: F403

# These imports may overwrite names from the import * above.
from ._aliases import *  # noqa: F403

__array_api_version__: Final = "2024.12"

# See the comment in the numpy __init__.py
__import__(__package__ + '.linalg')
__import__(__package__ + '.fft')



import pytest

import numpy as np
from numpy.linalg import lstsq
from numpy.testing import assert_allclose, assert_equal, assert_

from scipy.sparse import random_array, coo_array
from scipy.sparse.linalg import aslinearoperator
from scipy.optimize import lsq_linear
from scipy.optimize._minimize import Bounds


A = np.array([
    [0.171, -0.057],
    [-0.049, -0.248],
    [-0.166, 0.054],
])
b = np.array([0.074, 1.014, -0.383])


class BaseMixin:
    def setup_method(self):
        self.rnd = np.random.RandomState(0)

    def test_dense_no_bounds(self):
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, method=self.method, lsq_solver=lsq_solver)
            assert_allclose(res.x, lstsq(A, b, rcond=-1)[0])
            assert_allclose(res.x, res.unbounded_sol[0])

    def test_dense_bounds(self):
        # Solutions for comparison are taken from MATLAB.
        lb = np.array([-1, -10])
        ub = np.array([1, 0])
        unbounded_sol = lstsq(A, b, rcond=-1)[0]
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (lb, ub), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.x, lstsq(A, b, rcond=-1)[0])
            assert_allclose(res.unbounded_sol[0], unbounded_sol)

        lb = np.array([0.0, -np.inf])
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (lb, np.inf), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.x, np.array([0.0, -4.084174437334673]),
                            atol=1e-6)
            assert_allclose(res.unbounded_sol[0], unbounded_sol)

        lb = np.array([-1, 0])
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (lb, np.inf), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.x, np.array([0.448427311733504, 0]),
                            atol=1e-15)
            assert_allclose(res.unbounded_sol[0], unbounded_sol)

        ub = np.array([np.inf, -5])
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (-np.inf, ub), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.x, np.array([-0.105560998682388, -5]))
            assert_allclose(res.unbounded_sol[0], unbounded_sol)

        ub = np.array([-1, np.inf])
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (-np.inf, ub), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.x, np.array([-1, -4.181102129483254]))
            assert_allclose(res.unbounded_sol[0], unbounded_sol)

        lb = np.array([0, -4])
        ub = np.array([1, 0])
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (lb, ub), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.x, np.array([0.005236663400791, -4]))
            assert_allclose(res.unbounded_sol[0], unbounded_sol)

    def test_bounds_variants(self):
        x = np.array([1, 3])
        A = self.rnd.uniform(size=(2, 2))
        b = A@x
        lb = np.array([1, 1])
        ub = np.array([2, 2])
        bounds_old = (lb, ub)
        bounds_new = Bounds(lb, ub)
        res_old = lsq_linear(A, b, bounds_old)
        res_new = lsq_linear(A, b, bounds_new)
        assert not np.allclose(res_new.x, res_new.unbounded_sol[0])
        assert_allclose(res_old.x, res_new.x)

    def test_np_matrix(self):
        # gh-10711
        with np.testing.suppress_warnings() as sup:
            sup.filter(PendingDeprecationWarning)
            A = np.matrix([[20, -4, 0, 2, 3], [10, -2, 1, 0, -1]])
        k = np.array([20, 15])
        lsq_linear(A, k)

    def test_dense_rank_deficient(self):
        A = np.array([[-0.307, -0.184]])
        b = np.array([0.773])
        lb = [-0.1, -0.1]
        ub = [0.1, 0.1]
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (lb, ub), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.x, [-0.1, -0.1])
            assert_allclose(res.unbounded_sol[0], lstsq(A, b, rcond=-1)[0])

        A = np.array([
            [0.334, 0.668],
            [-0.516, -1.032],
            [0.192, 0.384],
        ])
        b = np.array([-1.436, 0.135, 0.909])
        lb = [0, -1]
        ub = [1, -0.5]
        for lsq_solver in self.lsq_solvers:
            res = lsq_linear(A, b, (lb, ub), method=self.method,
                             lsq_solver=lsq_solver)
            assert_allclose(res.optimality, 0, atol=1e-11)
            assert_allclose(res.unbounded_sol[0], lstsq(A, b, rcond=-1)[0])

    def test_full_result(self):
        lb = np.array([0, -4])
        ub = np.array([1, 0])
        res = lsq_linear(A, b, (lb, ub), method=self.method)

        assert_allclose(res.x, [0.005236663400791, -4])
        assert_allclose(res.unbounded_sol[0], lstsq(A, b, rcond=-1)[0])

        r = A.dot(res.x) - b
        assert_allclose(res.cost, 0.5 * np.dot(r, r))
        assert_allclose(res.fun, r)

        assert_allclose(res.optimality, 0.0, atol=1e-12)
        assert_equal(res.active_mask, [0, -1])
        assert_(res.nit < 15)
        assert_(res.status == 1 or res.status == 3)
        assert_(isinstance(res.message, str))
        assert_(res.success)

    # This is a test for issue #9982.
    def test_almost_singular(self):
        A = np.array(
            [[0.8854232310355122, 0.0365312146937765, 0.0365312146836789],
             [0.3742460132129041, 0.0130523214078376, 0.0130523214077873],
             [0.9680633871281361, 0.0319366128718639, 0.0319366128718388]])

        b = np.array(
            [0.0055029366538097, 0.0026677442422208, 0.0066612514782381])

        result = lsq_linear(A, b, method=self.method)
        assert_(result.cost < 1.1e-8)

    @pytest.mark.xslow
    def test_large_rank_deficient(self):
        np.random.seed(0)
        n, m = np.sort(np.random.randint(2, 1000, size=2))
        m *= 2   # make m >> n
        A = 1.0 * np.random.randint(-99, 99, size=[m, n])
        b = 1.0 * np.random.randint(-99, 99, size=[m])
        bounds = 1.0 * np.sort(np.random.randint(-99, 99, size=(2, n)), axis=0)
        bounds[1, :] += 1.0  # ensure up > lb

        # Make the A matrix strongly rank deficient by replicating some columns
        w = np.random.choice(n, n)  # Select random columns with duplicates
        A = A[:, w]

        x_bvls = lsq_linear(A, b, bounds=bounds, method='bvls').x
        x_trf = lsq_linear(A, b, bounds=bounds, method='trf').x

        cost_bvls = np.sum((A @ x_bvls - b)**2)
        cost_trf = np.sum((A @ x_trf - b)**2)

        assert_(abs(cost_bvls - cost_trf) < cost_trf*1e-10)

    def test_convergence_small_array(self):
        A = np.array([[49.0, 41.0, -32.0],
                      [-19.0, -32.0, -8.0],
                      [-13.0, 10.0, 69.0]])
        b = np.array([-41.0, -90.0, 47.0])
        bounds = np.array([[31.0, -44.0, 26.0],
                           [54.0, -32.0, 28.0]])

        x_bvls = lsq_linear(A, b, bounds=bounds, method='bvls').x
        x_trf = lsq_linear(A, b, bounds=bounds, method='trf').x

        cost_bvls = np.sum((A @ x_bvls - b)**2)
        cost_trf = np.sum((A @ x_trf - b)**2)

        assert_(abs(cost_bvls - cost_trf) < cost_trf*1e-10)


class SparseMixin:
    def test_sparse_and_LinearOperator(self):
        m = 5000
        n = 1000
        rng = np.random.RandomState(0)
        A = random_array((m, n), random_state=rng)
        b = rng.randn(m)
        res = lsq_linear(A, b)
        assert_allclose(res.optimality, 0, atol=1e-6)

        A = aslinearoperator(A)
        res = lsq_linear(A, b)
        assert_allclose(res.optimality, 0, atol=1e-6)

    @pytest.mark.fail_slow(10)
    def test_sparse_bounds(self):
        m = 5000
        n = 1000
        rng = np.random.RandomState(0)
        A = random_array((m, n), random_state=rng)
        b = rng.randn(m)
        lb = rng.randn(n)
        ub = lb + 1
        res = lsq_linear(A, b, (lb, ub))
        assert_allclose(res.optimality, 0.0, atol=1e-6)

        res = lsq_linear(A, b, (lb, ub), lsmr_tol=1e-13,
                         lsmr_maxiter=1500)
        assert_allclose(res.optimality, 0.0, atol=1e-6)

        res = lsq_linear(A, b, (lb, ub), lsmr_tol='auto')
        assert_allclose(res.optimality, 0.0, atol=1e-6)

    def test_sparse_ill_conditioned(self):
        # Sparse matrix with condition number of ~4 million
        data = np.array([1., 1., 1., 1. + 1e-6, 1.])
        row = np.array([0, 0, 1, 2, 2])
        col = np.array([0, 2, 1, 0, 2])
        A = coo_array((data, (row, col)), shape=(3, 3))

        # Get the exact solution
        exact_sol = lsq_linear(A.toarray(), b, lsq_solver='exact')

        # Default lsmr arguments should not fully converge the solution
        default_lsmr_sol = lsq_linear(A, b, lsq_solver='lsmr')
        with pytest.raises(AssertionError):
            assert_allclose(exact_sol.x, default_lsmr_sol.x)

        # By increasing the maximum lsmr iters, it will converge
        conv_lsmr = lsq_linear(A, b, lsq_solver='lsmr', lsmr_maxiter=10)
        assert_allclose(exact_sol.x, conv_lsmr.x)


class TestTRF(BaseMixin, SparseMixin):
    method = 'trf'
    lsq_solvers = ['exact', 'lsmr']


class TestBVLS(BaseMixin):
    method = 'bvls'
    lsq_solvers = ['exact']


class TestErrorChecking:
    def test_option_lsmr_tol(self):
        # Should work with a positive float, string equal to 'auto', or None
        _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_tol=1e-2)
        _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_tol='auto')
        _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_tol=None)

        # Should raise error with negative float, strings
        # other than 'auto', and integers
        err_message = "`lsmr_tol` must be None, 'auto', or positive float."
        with pytest.raises(ValueError, match=err_message):
            _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_tol=-0.1)
        with pytest.raises(ValueError, match=err_message):
            _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_tol='foo')
        with pytest.raises(ValueError, match=err_message):
            _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_tol=1)

    def test_option_lsmr_maxiter(self):
        # Should work with positive integers or None
        _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_maxiter=1)
        _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_maxiter=None)

        # Should raise error with 0 or negative max iter
        err_message = "`lsmr_maxiter` must be None or positive integer."
        with pytest.raises(ValueError, match=err_message):
            _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_maxiter=0)
        with pytest.raises(ValueError, match=err_message):
            _ = lsq_linear(A, b, lsq_solver='lsmr', lsmr_maxiter=-1)



import numpy as np
import matplotlib.pyplot as plt
from scipy import stats

npoints = 20  # number of integer support points of the distribution minus 1
npointsh = npoints // 2
npointsf = float(npoints)
nbound = 4  # bounds for the truncated normal
normbound = (1 + 1/npointsf) * nbound  # actual bounds of truncated normal
grid = np.arange(-npointsh, npointsh+2, 1)  # integer grid
gridlimitsnorm = (grid-0.5) / npointsh * nbound  # bin limits for the truncnorm
gridlimits = grid - 0.5
grid = grid[:-1]
probs = np.diff(stats.truncnorm.cdf(gridlimitsnorm, -normbound, normbound))
gridint = grid
normdiscrete = stats.rv_discrete(
                        values=(gridint, np.round(probs, decimals=7)),
                        name='normdiscrete')

n_sample = 500
rng = np.random.default_rng()
rvs = normdiscrete.rvs(size=n_sample, random_state=rng)
f, l = np.histogram(rvs, bins=gridlimits)
sfreq = np.vstack([gridint, f, probs*n_sample]).T
fs = sfreq[:,1] / float(n_sample)
ft = sfreq[:,2] / float(n_sample)
nd_std = np.sqrt(normdiscrete.stats(moments='v'))

ind = gridint  # the x locations for the groups
width = 0.35       # the width of the bars

plt.subplot(111)
rects1 = plt.bar(ind, ft, width, color='b')
rects2 = plt.bar(ind+width, fs, width, color='r')
normline = plt.plot(ind+width/2.0, stats.norm.pdf(ind, scale=nd_std),
                    color='b')

plt.ylabel('Frequency')
plt.title('Frequency and Probability of normdiscrete')
plt.xticks(ind+width, ind)
plt.legend((rects1[0], rects2[0]), ('true', 'sample'))

plt.show()



cdef extern from "limits.h":
    unsigned long ULONG_MAX


def _comb_int(N, k):
    # Fast path with machine integers
    try:
        r = _comb_int_long(N, k)
        if r != 0:
            return r
    except (OverflowError, TypeError):
        pass

    # Fallback
    N = int(N)
    k = int(k)

    if k > N or N < 0 or k < 0:
        return 0

    M = N + 1
    nterms = min(k, N - k)

    numerator = 1
    denominator = 1
    for j in range(1, nterms + 1):
        numerator *= M - j
        denominator *= j

    return numerator // denominator


cdef unsigned long _comb_int_long(unsigned long N, unsigned long k) noexcept:
    """
    Compute binom(N, k) for integers.
    Returns 0 if error/overflow encountered.
    """
    cdef unsigned long val, j, M, nterms

    if k > N or N == ULONG_MAX:
        return 0

    M = N + 1
    nterms = min(k, N - k)

    val = 1

    for j in range(1, nterms + 1):
        # Overflow check
        if val > ULONG_MAX // (M - j):
            return 0

        val *= M - j
        val //= j

    return val



import os
import pytest
import sys

@pytest.fixture(scope='function')
def minimize_with_debugging():
    # This is a hack to force us to use DEBUGGING for a test, so that we get
    # code coverage. We definitely don't want to do this for the pycutest tests,
    # they are slow enough with USE_NAIVE_MATH set to True.
    os.environ['PRIMA_DEBUGGING'] = "True"
    if 'pyprima' in sys.modules:
        modules_to_delete = [m for m in sys.modules if 'pyprima' in m]
        for m in modules_to_delete:
            del sys.modules[m]
    # We have to reimport minimize here, if the tests try to use the minimize that
    # was imported at the top of the file, it will be the existing object which
    # captured DEBUGGING=False.
    from pyprima import minimize
    yield minimize
    del os.environ['PRIMA_DEBUGGING']


# -*- python -*- or near enough

import zlib

from cpython cimport PyBytes_AS_STRING, PyBytes_Size

from ._pyalloc cimport pyalloc_v

from libc.string cimport memcpy

cdef extern from "Python.h":
    void *PyCObject_Import(char *, char *) except NULL
    ctypedef struct PyTypeObject:
        pass
    ctypedef struct PyObject:
        pass
    ctypedef struct FILE


DEF _BLOCK_SIZE = 131072

BLOCK_SIZE = _BLOCK_SIZE  # public

cdef class GenericStream:

    def __init__(self, fobj):
        self.fobj = fobj

    cpdef int seek(self, long int offset, int whence=0) except -1:
        self.fobj.seek(offset, whence)
        return 0

    cpdef long int tell(self) except -1:
        return self.fobj.tell()

    def read(self, n_bytes):
        return self.fobj.read(n_bytes)

    cpdef int all_data_read(self) except *:
        return 1

    cdef int read_into(self, void *buf, size_t n) except -1:
        """ Read n bytes from stream into pre-allocated buffer `buf`
        """
        cdef char *p
        cdef size_t read_size, count

        # Read data to buf in _BLOCK_SIZE blocks
        count = 0
        p = <char*>buf
        while count < n:
            read_size = min(n - count, _BLOCK_SIZE)
            data = self.fobj.read(read_size)
            read_size = len(data)
            if read_size == 0:
                break
            memcpy(p, <const char*>data, read_size)
            p += read_size
            count += read_size

        if count != n:
            raise OSError('could not read bytes')
        return 0

    cdef object read_string(self, size_t n, void **pp, int copy=True):
        """Make new memory, wrap with object"""
        if copy != True:
            data = self.fobj.read(n)
            if PyBytes_Size(data) != n:
                raise OSError('could not read bytes')
            pp[0] = <void*>PyBytes_AS_STRING(data)
            return data

        cdef object d_copy = pyalloc_v(n, pp)
        self.read_into(pp[0], n)
        return d_copy


cdef class ZlibInputStream(GenericStream):
    """
    File-like object uncompressing bytes from a zlib compressed stream.

    Parameters
    ----------
    stream : file-like
        Stream to read compressed data from.
    max_length : int
        Maximum number of bytes to read from the stream.

    Notes
    -----
    Some matlab files contain zlib streams without valid Z_STREAM_END
    termination.  To get round this, we use the decompressobj object, that
    allows you to decode an incomplete stream.  See discussion at
    https://bugs.python.org/issue8672

    """

    cdef ssize_t _max_length
    cdef object _decompressor
    cdef bytes _buffer
    cdef size_t _buffer_size
    cdef size_t _buffer_position
    cdef size_t _total_position
    cdef size_t _read_bytes

    def __init__(self, fobj, ssize_t max_length):
        self.fobj = fobj

        self._max_length = max_length
        self._decompressor = zlib.decompressobj()
        self._buffer = b''
        self._buffer_size = 0
        self._buffer_position = 0
        self._total_position = 0
        self._read_bytes = 0

    cdef inline void _fill_buffer(self) except *:
        cdef size_t read_size
        cdef bytes block

        if self._buffer_position < self._buffer_size:
            return

        read_size = min(_BLOCK_SIZE, self._max_length - self._read_bytes)

        block = self.fobj.read(read_size)
        self._read_bytes += len(block)

        self._buffer_position = 0
        if not block:
            self._buffer = self._decompressor.flush()
        else:
            self._buffer = self._decompressor.decompress(block)
        self._buffer_size = len(self._buffer)

    cdef int read_into(self, void *buf, size_t n) except -1:
        """Read n bytes from stream into pre-allocated buffer `buf`
        """
        cdef char *dstp
        cdef char *srcp
        cdef size_t count, size

        dstp = <char*>buf
        count = 0
        while count < n:
            self._fill_buffer()
            if self._buffer_size == 0:
                break

            srcp = <char*>self._buffer
            srcp += self._buffer_position

            size = min(n - count, self._buffer_size - self._buffer_position)
            memcpy(dstp, srcp, size)

            count += size
            dstp += size
            self._buffer_position += size

        self._total_position += count

        if count != n:
            raise OSError('could not read bytes')

        return 0

    cdef object read_string(self, size_t n, void **pp, int copy=True):
        """Make new memory, wrap with object"""
        cdef object d_copy = pyalloc_v(n, pp)
        self.read_into(pp[0], n)
        return d_copy

    def read(self, n_bytes):
        cdef void *p
        return self.read_string(n_bytes, &p)

    cpdef int all_data_read(self) except *:
        if self._read_bytes < self._max_length:
            # we might still have checksum bytes to read
            self._fill_buffer()
        return (self._max_length == self._read_bytes) and \
               (self._buffer_size == self._buffer_position)

    cpdef long int tell(self) except -1:
        if self._total_position == -1:
            raise OSError("Invalid file position.")
        return self._total_position

    cpdef int seek(self, long int offset, int whence=0) except -1:
        cdef ssize_t new_pos, size
        if whence == 1:
            new_pos = <ssize_t>self._total_position + offset
        elif whence == 0:
            new_pos = offset
        elif whence == 2:
            raise OSError("Zlib stream cannot seek from file end")
        else:
            raise ValueError("Invalid value for whence")

        if new_pos < self._total_position:
            raise OSError("Zlib stream cannot seek backwards")

        while self._total_position < new_pos:
            self._fill_buffer()
            if self._buffer_size == 0:
                break

            size = min(new_pos - self._total_position,
                       self._buffer_size - self._buffer_position)

            self._total_position += size
            self._buffer_position += size

        return 0


def _read_into(GenericStream st, size_t n):
    # for testing only.  Use st.read instead
    cdef char * d_ptr
    # use bytearray because bytes() is immutable
    my_str = bytearray(b' ' * n)
    d_ptr = my_str
    st.read_into(d_ptr, n)
    return bytes(my_str)


def _read_string(GenericStream st, size_t n):
    # for testing only.  Use st.read instead
    cdef void *d_ptr
    _obj = st.read_string(n, &d_ptr, True)
    # use bytearray because bytes() is immutable
    my_str = bytearray(b'A' * n)
    cdef char *mys_ptr = my_str
    memcpy(mys_ptr, d_ptr, n)
    return bytes(my_str)


cpdef GenericStream make_stream(object fobj):
    """ Make stream of correct type for file-like `fobj`
    """
    if isinstance(fobj, GenericStream):
        return fobj
    return GenericStream(fobj)



from contextlib import contextmanager

import numpy as np
import pytest

try:
    import dask
    import dask.array as da
except ImportError:
    pytestmark = pytest.skip(allow_module_level=True, reason="dask not found")

from array_api_compat import array_namespace


@pytest.fixture
def xp():
    """Fixture returning the wrapped dask namespace"""
    return array_namespace(da.empty(0))


@contextmanager
def assert_no_compute():
    """
    Context manager that raises if at any point inside it anything calls compute()
    or persist(), e.g. as it can be triggered implicitly by __bool__, __array__, etc.
    """

    def get(dsk, *args, **kwargs):
        raise AssertionError("Called compute() or persist()")

    with dask.config.set(scheduler=get):
        yield


def test_assert_no_compute():
    """Test the assert_no_compute context manager"""
    a = da.asarray(True)
    with pytest.raises(AssertionError, match="Called compute"):
        with assert_no_compute():
            bool(a)

    # Exiting the context manager restores the original scheduler
    assert bool(a) is True


# Test no_compute for functions that use generic _aliases with xp=np


def test_unary_ops_no_compute(xp):
    with assert_no_compute():
        a = xp.asarray([1.5, -1.5])
        xp.ceil(a)
        xp.floor(a)
        xp.trunc(a)
        xp.sign(a)


def test_matmul_tensordot_no_compute(xp):
    A = da.ones((4, 4), chunks=2)
    B = da.zeros((4, 4), chunks=2)
    with assert_no_compute():
        xp.matmul(A, B)
        xp.tensordot(A, B)


# Test no_compute for functions that are fully bespoke for dask


def test_asarray_no_compute(xp):
    with assert_no_compute():
        a = xp.arange(10)
        xp.asarray(a)
        xp.asarray(a, dtype=np.int16)
        xp.asarray(a, dtype=a.dtype)
        xp.asarray(a, copy=True)
        xp.asarray(a, copy=True, dtype=np.int16)
        xp.asarray(a, copy=True, dtype=a.dtype)
        xp.asarray(a, copy=False)
        xp.asarray(a, copy=False, dtype=a.dtype)


@pytest.mark.parametrize("copy", [True, False])
def test_astype_no_compute(xp, copy):
    with assert_no_compute():
        a = xp.arange(10)
        xp.astype(a, np.int16, copy=copy)
        xp.astype(a, a.dtype, copy=copy)


def test_clip_no_compute(xp):
    with assert_no_compute():
        a = xp.arange(10)
        xp.clip(a)
        xp.clip(a, 1)
        xp.clip(a, 1, 8)


@pytest.mark.parametrize("chunks", (5, 10))
def test_sort_argsort_nocompute(xp, chunks):
    with assert_no_compute():
        a = xp.arange(10, chunks=chunks)
        xp.sort(a)
        xp.argsort(a)


def test_generators_are_lazy(xp):
    """
    Test that generator functions are fully lazy, e.g. that
    da.ones(n) is not implemented as da.asarray(np.ones(n))
    """
    size = 100_000_000_000  # 800 GB
    chunks = size // 10  # 10x 80 GB chunks

    with assert_no_compute():
        xp.zeros(size, chunks=chunks)
        xp.ones(size, chunks=chunks)
        xp.empty(size, chunks=chunks)
        xp.full(size, fill_value=123, chunks=chunks)
        a = xp.arange(size, chunks=chunks)
        xp.zeros_like(a)
        xp.ones_like(a)
        xp.empty_like(a)
        xp.full_like(a, fill_value=123)


@pytest.mark.parametrize("axis", [0, 1])
@pytest.mark.parametrize("func", ["sort", "argsort"])
def test_sort_argsort_chunks(xp, func, axis):
    """Test that sort and argsort are functionally correct when
    the array is chunked along the sort axis, e.g. the sort is
    not just local to each chunk.
    """
    a = da.random.random((10, 10), chunks=(5, 5))
    actual = getattr(xp, func)(a, axis=axis)
    expect = getattr(np, func)(a.compute(), axis=axis)
    np.testing.assert_array_equal(actual, expect)


@pytest.mark.parametrize(
    "shape,chunks",
    [
        # 3 GiB; 128 MiB per chunk; must rechunk before sorting.
        # Sort chunks can be 128 MiB each; no need for final rechunk.
        ((20_000, 20_000), "auto"),
        # 3 GiB; 128 MiB per chunk; must rechunk before sorting.
        # Must sort on two 1.5 GiB chunks; benefits from final rechunk.
        ((2, 2**30 * 3 // 16), "auto"),
        # 3 GiB; 1.5 GiB per chunk; no need to rechunk before sorting.
        # Surely the user must know what they're doing, so don't
        # perform the final rechunk.
        ((2, 2**30 * 3 // 16), (1, -1)),
    ],
)
@pytest.mark.parametrize("func", ["sort", "argsort"])
def test_sort_argsort_chunk_size(xp, func, shape, chunks):
    """
    Test that sort and argsort produce reasonably-sized chunks
    in the output array, even if they had to go through a singular
    huge one to perform the operation.
    """
    a = da.random.random(shape, chunks=chunks)
    b = getattr(xp, func)(a)
    max_chunk_size = max(b.chunks[0]) * max(b.chunks[1]) * b.dtype.itemsize
    assert (
        max_chunk_size <= 128 * 1024 * 1024  # 128 MiB
        or b.chunks == a.chunks
    )


@pytest.mark.parametrize("func", ["sort", "argsort"])
def test_sort_argsort_meta(xp, func):
    """Test meta-namespace other than numpy"""
    mxp = pytest.importorskip("array_api_strict")
    typ = type(mxp.asarray(0))
    a = da.random.random(10)
    b = a.map_blocks(mxp.asarray)
    assert isinstance(b._meta, typ)
    c = getattr(xp, func)(b)
    assert isinstance(c._meta, typ)
    d = c.compute()
    # Note: np.sort(array_api_strict.asarray(0)) would return a numpy array
    assert isinstance(d, typ)
    np.testing.assert_array_equal(d, getattr(np, func)(a.compute()))



"""
Spherical Voronoi Cython Code

.. versionadded:: 0.19.0

"""
#
# Copyright (C)  Tyler Reddy, 2016
#
# Distributed under the same BSD license as Scipy.
#

import numpy as np
cimport numpy as np
cimport cython

np.import_array()

__all__ = ['sort_vertices_of_regions']


@cython.boundscheck(False)
def sort_vertices_of_regions(const int[:,::1] simplices, list regions):
    cdef np.npy_intp n, k, s, i, max_len
    cdef np.npy_intp num_regions = len(regions)
    cdef np.npy_intp current_simplex, current_vertex
    cdef np.npy_intp remaining_size
    cdef np.npy_intp[:] remaining
    cdef np.ndarray[np.intp_t, ndim=1] sorted_vertices

    max_len = 0
    for region in regions:
        max_len = max(max_len, len(region))
    sorted_vertices = np.empty(max_len, dtype=np.intp)

    for n in range(num_regions):
        remaining = np.asarray(regions[n][:])
        remaining_size = remaining.shape[0]
        current_simplex = remaining[0]
        for i in range(3):
            k = simplices[current_simplex, i]
            if k != n:
                current_vertex = k
                break
        for k in range(remaining_size):
            sorted_vertices[k] = current_simplex
            for i in range(remaining_size):
                if remaining[i] == sorted_vertices[k]:
                    continue
                s = remaining[i]
                for j in range(3):
                    if current_vertex == simplices[s, j]:
                        current_simplex = s
            for i in range(3):
                s = simplices[current_simplex, i]
                if s != n and s != current_vertex:
                    current_vertex = s
                    break
        regions[n] = list(sorted_vertices[:remaining_size])



# cython: language_level=3
# cython: boundscheck=False
# cython: wraparound=False
# cython: cdivision=True
# cython: cpow=True


import numpy as np
cimport numpy as np
from libc.math cimport fabs, sqrt, pow

np.import_array()

cdef extern from "<thread>" namespace "std" nogil:
    cdef cppclass thread:
        thread()
        # Since we are using the C++ thread constructor, Cython forces us
        # to define how many arguments will take the constructor. Here,
        # `_cy_wrapper_van_der_corput_scrambled` takes 7 arguments thus
        # the `A, B, C, D, E, F, G` inside the constructor. Since the other
        # threaded loops take less arguments, we need to find a workaround
        # to pass 7 arguments to the functions. Here we use `_`.
        void thread[A, B, C, D, E, F, G](A, B, C, D, E, F, G)
        void join()

cdef extern from "<mutex>" namespace "std" nogil:
    cdef cppclass mutex:
        void lock()
        void unlock()

cdef extern from "<functional>" namespace "std" nogil:
    cdef cppclass reference_wrapper[T]:
        pass
    cdef reference_wrapper[T] ref[T](T&)

from libcpp.vector cimport vector


cdef mutex threaded_sum_mutex

def _cy_wrapper_centered_discrepancy(const double[:, ::1] sample, bint iterative,
                                     workers):
    return centered_discrepancy(sample, iterative, workers)


def _cy_wrapper_wrap_around_discrepancy(const double[:, ::1] sample,
                                        bint iterative, workers):
    return wrap_around_discrepancy(sample, iterative, workers)


def _cy_wrapper_mixture_discrepancy(const double[:, ::1] sample,
                                    bint iterative, workers):
    return mixture_discrepancy(sample, iterative, workers)


def _cy_wrapper_l2_star_discrepancy(const double[:, ::1] sample,
                                    bint iterative, workers):
    return l2_star_discrepancy(sample, iterative, workers)


cdef double centered_discrepancy(const double[:, ::1] sample_view,
                                 bint iterative, unsigned int workers) noexcept nogil:
    cdef:
        Py_ssize_t n = sample_view.shape[0]
        Py_ssize_t d = sample_view.shape[1]
        Py_ssize_t i = 0, j = 0
        double prod, disc1 = 0

    for i in range(n):
        prod = 1
        for j in range(d):
            prod *= (
                1 + 0.5 * fabs(sample_view[i, j] - 0.5) - 0.5
                * fabs(sample_view[i, j] - 0.5) ** 2
            )
        disc1 += prod

    cdef double disc2 = threaded_loops(centered_discrepancy_loop, sample_view,
                                       workers)

    if iterative:
        n += 1

    return ((13.0 / 12.0) ** d - 2.0 / n * disc1
            + 1.0 / (n ** 2) * disc2)


cdef double centered_discrepancy_loop(const double[:, ::1] sample_view,
                                      Py_ssize_t istart, Py_ssize_t istop) noexcept nogil:

    cdef:
        Py_ssize_t i, j, k
        double prod, disc2 = 0

    for i in range(istart, istop):
        for j in range(sample_view.shape[0]):
            prod = 1
            for k in range(sample_view.shape[1]):
                prod *= (
                    1 + 0.5 * fabs(sample_view[i, k] - 0.5)
                    + 0.5 * fabs(sample_view[j, k] - 0.5)
                    - 0.5 * fabs(sample_view[i, k] - sample_view[j, k])
                )
            disc2 += prod

    return disc2


cdef double wrap_around_discrepancy(const double[:, ::1] sample_view,
                                    bint iterative, unsigned int workers) noexcept nogil:
    cdef:
        Py_ssize_t n = sample_view.shape[0]
        Py_ssize_t d = sample_view.shape[1]
        double disc

    disc = threaded_loops(wrap_around_loop, sample_view,
                          workers)

    if iterative:
        n += 1

    return - (4.0 / 3.0) ** d + 1.0 / (n ** 2) * disc


cdef double wrap_around_loop(const double[:, ::1] sample_view,
                             Py_ssize_t istart, Py_ssize_t istop) noexcept nogil:

    cdef:
        Py_ssize_t i, j, k
        double prod, disc = 0

    for i in range(istart, istop):
        for j in range(sample_view.shape[0]):
            prod = 1
            for k in range(sample_view.shape[1]):
                x_kikj = fabs(sample_view[i, k] - sample_view[j, k])
                prod *= 3.0 / 2.0 - x_kikj + x_kikj ** 2
            disc += prod

    return disc


cdef double mixture_discrepancy(const double[:, ::1] sample_view,
                                bint iterative, unsigned int workers) noexcept nogil:
    cdef:
        Py_ssize_t n = sample_view.shape[0]
        Py_ssize_t d = sample_view.shape[1]
        Py_ssize_t i = 0, j = 0
        double prod = 1, disc = 0, disc1 = 0

    for i in range(n):
        for j in range(d):
            prod *= (
                5.0 / 3.0 - 0.25 * fabs(sample_view[i, j] - 0.5)
                - 0.25 * fabs(sample_view[i, j] - 0.5) ** 2
            )
        disc1 += prod
        prod = 1

    cdef double disc2 = threaded_loops(mixture_loop, sample_view, workers)

    if iterative:
        n += 1

    disc = (19.0 / 12.0) ** d
    disc1 = 2.0 / n * disc1
    disc2 = 1.0 / (n ** 2) * disc2

    return disc - disc1 + disc2


cdef double mixture_loop(const double[:, ::1] sample_view, Py_ssize_t istart,
                         Py_ssize_t istop) noexcept nogil:

    cdef:
        Py_ssize_t i, j, k
        double prod, disc2 = 0

    for i in range(istart, istop):
        for j in range(sample_view.shape[0]):
            prod = 1
            for k in range(sample_view.shape[1]):
                prod *= (15.0 / 8.0
                         - 0.25 * fabs(sample_view[i, k] - 0.5)
                         - 0.25 * fabs(sample_view[j, k] - 0.5)
                         - 3.0 / 4.0 * fabs(sample_view[i, k]
                                            - sample_view[j, k])
                         + 0.5
                         * fabs(sample_view[i, k] - sample_view[j, k]) ** 2)
            disc2 += prod

    return disc2


cdef double l2_star_discrepancy(const double[:, ::1] sample_view,
                                bint iterative, unsigned int workers) noexcept nogil:
    cdef:
        Py_ssize_t n = sample_view.shape[0]
        Py_ssize_t d = sample_view.shape[1]
        Py_ssize_t i = 0, j = 0
        double prod = 1, disc1 = 0

    for i in range(n):
        for j in range(d):
            prod *= 1 - sample_view[i, j] ** 2

        disc1 += prod
        prod = 1

    cdef double disc2 = threaded_loops(l2_star_loop, sample_view, workers)

    if iterative:
        n += 1

    cdef double one_div_n = <double> 1 / n
    return sqrt(
        pow(3, -d) - one_div_n * pow(2, 1 - d) * disc1 + 1 / pow(n, 2) * disc2
    )


cdef double l2_star_loop(const double[:, ::1] sample_view, Py_ssize_t istart,
                         Py_ssize_t istop) noexcept nogil:

    cdef:
        Py_ssize_t i, j, k
        double prod = 1, disc2 = 0, tmp_sum = 0

    for i in range(istart, istop):
        for j in range(sample_view.shape[0]):
            prod = 1
            for k in range(sample_view.shape[1]):
                prod *= (
                    1 - max(sample_view[i, k], sample_view[j, k])
                )
            tmp_sum += prod

        disc2 += tmp_sum
        tmp_sum = 0

    return disc2


def _cy_wrapper_update_discrepancy(const double[::1] x_new_view,
                                   const double[:, ::1] sample_view,
                                   double initial_disc):
    return c_update_discrepancy(x_new_view, sample_view, initial_disc)


cdef double c_update_discrepancy(const double[::1] x_new_view,
                                 const double[:, ::1] sample_view,
                                 double initial_disc) noexcept:
    cdef:
        Py_ssize_t n = sample_view.shape[0] + 1
        Py_ssize_t d = sample_view.shape[1]
        Py_ssize_t i = 0, j = 0
        double prod = 1
        double  disc1 = 0, disc2 = 0, disc3 = 0
        double[::1] abs_ = np.empty(d, dtype=np.float64)


    # derivation from P.T. Roy (@tupui)
    for i in range(d):
        abs_[i] = fabs(x_new_view[i] - 0.5)
        prod *= (
            1 + 0.5 * abs_[i]
            - 0.5 * pow(abs_[i], 2)
        )

    disc1 = (- 2 / <double> n) * prod

    prod = 1
    for i in range(n - 1):
        for j in range(d):
            prod *= (
                1 + 0.5 * abs_[j]
                + 0.5 * fabs(sample_view[i, j] - 0.5)
                - 0.5 * fabs(x_new_view[j] - sample_view[i, j])
            )
        disc2 += prod
        prod = 1

    disc2 *= 2 / pow(n, 2)

    for i in range(d):
        prod *= 1 + abs_[i]

    disc3 = 1 / pow(n, 2) * prod

    return initial_disc + disc1 + disc2 + disc3


ctypedef double (*func_type)(const double[:, ::1], Py_ssize_t,
                             Py_ssize_t) noexcept nogil


cdef double threaded_loops(func_type loop_func,
                           const double[:, ::1] sample_view,
                           unsigned int workers) noexcept nogil:
    cdef:
        Py_ssize_t n = sample_view.shape[0]
        double disc2 = 0

    if workers <= 1:
        return loop_func(sample_view, 0, n)

    cdef:
        vector[thread] threads
        unsigned int tid
        Py_ssize_t istart, istop

    for tid in range(workers):
        istart = <Py_ssize_t> (n / workers * tid)
        istop = <Py_ssize_t> (
            n / workers * (tid + 1)) if tid < workers - 1 else n
        threads.push_back(
            thread(one_thread_loop, loop_func, ref(disc2),
                   sample_view, istart, istop, None)
        )

    for tid in range(workers):
        threads[tid].join()

    return disc2


cdef void one_thread_loop(func_type loop_func,
                          double& disc,
                          const double[:, ::1] sample_view,
                          Py_ssize_t istart,
                          Py_ssize_t istop,
                          _) noexcept nogil:

    cdef double tmp = loop_func(sample_view, istart, istop)

    threaded_sum_mutex.lock()
    (&disc)[0] += tmp # workaround to "disc += tmp", see cython issue #1863
    threaded_sum_mutex.unlock()


def _cy_van_der_corput(Py_ssize_t n,
                       long base,
                       long start_index,
                       unsigned int workers):
    sequence = np.zeros(n, dtype=np.double)

    cdef:
        double[::1] sequence_view = sequence
        vector[thread] threads
        unsigned int tid
        Py_ssize_t istart, istop

    if workers <= 1:
        _cy_van_der_corput_threaded_loop(0, n, base, start_index,
                                         sequence_view, None)
        return sequence

    for tid in range(workers):
        istart = <Py_ssize_t> (n / workers * tid)
        istop = <Py_ssize_t> (
                n / workers * (tid + 1)) if tid < workers - 1 else n
        threads.push_back(
            thread(_cy_van_der_corput_threaded_loop, istart, istop, base,
                   start_index, sequence_view, None)
        )

    for tid in range(workers):
        threads[tid].join()

    return sequence


cdef _cy_van_der_corput_threaded_loop(Py_ssize_t istart,
                                      Py_ssize_t istop,
                                      long base,
                                      long start_index,
                                      double[::1] sequence_view,
                                      _):
    cdef:
        long quotient, remainder
        Py_ssize_t i
        double b2r

    for i in range(istart, istop):
        quotient = start_index + i
        b2r = 1.0 / base
        while quotient > 0:
            remainder = quotient % base
            sequence_view[i] += remainder * b2r
            b2r /= base
            quotient //= base


def _cy_van_der_corput_scrambled(Py_ssize_t n,
                                 long base,
                                 long start_index,
                                 const np.int64_t[:,::1] permutations,
                                 unsigned int workers):
    sequence = np.zeros(n)

    cdef:
        double[::1] sequence_view = sequence
        vector[thread] threads
        unsigned int tid
        Py_ssize_t istart, istop

    if workers <= 1:
        _cy_van_der_corput_scrambled_loop(0, n, base, start_index,
                                          permutations, sequence_view)
        return sequence

    for tid in range(workers):
        istart = <Py_ssize_t> (n / workers * tid)
        istop = <Py_ssize_t> (
                n / workers * (tid + 1)) if tid < workers - 1 else n
        threads.push_back(
            thread(_cy_van_der_corput_scrambled_loop, istart, istop, base,
                   start_index, permutations, sequence_view)
        )

    for tid in range(workers):
        threads[tid].join()

    return sequence


cdef _cy_van_der_corput_scrambled_loop(Py_ssize_t istart,
                                       Py_ssize_t istop,
                                       long base,
                                       long start_index,
                                       const np.int64_t[:,::1] permutations,
                                       double[::1] sequence_view):

    cdef:
        long i, j, quotient, remainder
        double b2r

    for i in range(istart, istop):
        quotient = start_index + i
        b2r = 1.0 / base
        for j in range(permutations.shape[0]):
            remainder = quotient % base
            remainder = permutations[j, remainder]
            sequence_view[i] += remainder * b2r
            b2r /= base
            quotient //= base



"""
Test that files that define __all__ aren't missing any exports.

You can add names that shouldn't be exported to _all_ignore, like

_all_ignore = ['sys']

This is preferable to del-ing the names as this will break any name that is
used inside of a function. Note that names starting with an underscore are automatically ignored.
"""


import sys

from ._helpers import import_, wrapped_libraries

import pytest
import typing

TYPING_NAMES = frozenset((
    "Array",
    "Device",
    "DType",
    "Namespace",
    "NestedSequence",
    "SupportsBufferProtocol",
))

@pytest.mark.parametrize("library", ["common"] + wrapped_libraries)
def test_all(library):
    if library == "common":
        import array_api_compat.common  # noqa: F401
    else:
        import_(library, wrapper=True)

    # NB: iterate over a copy to avoid a "dictionary size changed" error
    for mod_name in sys.modules.copy():
        if not mod_name.startswith('array_api_compat.' + library):
            continue

        module = sys.modules[mod_name]

        # TODO: We should define __all__ in the __init__.py files and test it
        # there too.
        if not hasattr(module, '__all__'):
            continue

        dir_names = [n for n in dir(module) if not n.startswith('_')]
        if '__array_namespace_info__' in dir(module):
            dir_names.append('__array_namespace_info__')
        ignore_all_names = set(getattr(module, '_all_ignore', ()))
        ignore_all_names |= set(dir(typing))
        ignore_all_names |= {"annotations"}
        if not module.__name__.endswith("._typing"):
            ignore_all_names |= TYPING_NAMES
        dir_names = set(dir_names) - set(ignore_all_names)
        all_names = module.__all__

        if set(dir_names) != set(all_names):
            extra_dir = set(dir_names) - set(all_names)
            extra_all = set(all_names) - set(dir_names)
            assert not extra_dir, f"Some dir() names not included in __all__ for {mod_name}: {extra_dir}"
            assert not extra_all, f"Some __all__ names not in dir() for {mod_name}: {extra_all}"



import numpy as np
import pytest
from scipy.optimize import Bounds, LinearConstraint, NonlinearConstraint
from scipy.optimize._minimize import standardize_constraints

from ..problem import (
    ObjectiveFunction,
    BoundConstraints,
    LinearConstraints,
    NonlinearConstraints,
    Problem,
)
from ..settings import PRINT_OPTIONS
from ..utils import CallbackSuccess


class BaseTest:

    @staticmethod
    def rosen(x, c=100.0):
        x = np.asarray(x)
        return np.sum(c * (x[1:] - x[:-1] ** 2.0) ** 2.0
                      + (1.0 - x[:-1]) ** 2.0)

    class Rosen:

        def __call__(self, x):
            return BaseTest.rosen(x)


class TestObjectiveFunction(BaseTest):

    def test_simple(self, capsys):
        obj = ObjectiveFunction(self.rosen, False, True)
        x = [1.5, 1.5]
        assert obj.n_eval == 0
        assert obj(x) == self.rosen(x)
        assert obj.n_eval == 1
        assert obj.name == "rosen"
        captured = capsys.readouterr()
        assert captured.out == ""

    def test_args(self):
        obj = ObjectiveFunction(self.rosen, False, True, 2.0)
        x = [1.5, 1.5]
        assert obj(x) == self.rosen(x, 2.0)

    def test_none(self):
        obj = ObjectiveFunction(None, False, True)
        assert obj([1.5, 1.5]) == 0.0
        assert obj.n_eval == 0
        assert obj.name == ""

    def test_wrapper(self):
        obj = ObjectiveFunction(self.Rosen(), False, True)
        assert obj.name == "fun"

    def test_verbose(self, capsys):
        obj = ObjectiveFunction(self.rosen, True, True)
        x = np.array([1.5, 1.5])
        obj(x)
        captured = capsys.readouterr()
        with np.printoptions(**PRINT_OPTIONS):
            assert captured.out == f"rosen({x}) = {self.rosen(x)}\n"


class TestBoundConstraints:

    def test_simple(self):
        bounds = Bounds([0.0, 0.0], [1.0, 1.0])
        constraints = BoundConstraints(bounds)
        np.testing.assert_array_equal(constraints.xl, bounds.lb)
        np.testing.assert_array_equal(constraints.xu, bounds.ub)
        assert constraints.m == 4
        assert constraints.is_feasible
        assert constraints.maxcv([0.5, 0.5]) == 0.0
        assert constraints.maxcv(constraints.xl) == 0.0
        assert constraints.maxcv(constraints.xu) == 0.0
        x = [2.0, 2.0]
        assert np.all(constraints.project(x) >= constraints.xl)
        assert np.all(constraints.project(x) <= constraints.xu)

    def test_nan(self):
        bounds = Bounds([np.nan, 0.0], [1.0, 1.0])
        constraints = BoundConstraints(bounds)
        np.testing.assert_array_equal(constraints.xl, [-np.inf, 0.0])
        assert constraints.m == 3
        bounds = Bounds([0.0, 0.0], [1.0, np.nan])
        constraints = BoundConstraints(bounds)
        np.testing.assert_array_equal(constraints.xu, [1.0, np.inf])
        assert constraints.m == 3

    def test_infeasible(self):
        bounds = Bounds([2.0, 0.0], [1.0, 1.0])
        constraints = BoundConstraints(bounds)
        assert not constraints.is_feasible
        x = [2.0, 2.0]
        np.testing.assert_array_equal(constraints.project(x), x)
        np.testing.assert_allclose(constraints.maxcv(x), 1.0, atol=1e-15)


class TestLinearConstraints:

    def test_simple(self):
        linear_constraints = [
            LinearConstraint([[1.0, 1.0]], [0.0], [1.0]),
            LinearConstraint([[2.0, 1.0]], [1.0], [1.0]),
        ]
        constraints = LinearConstraints(linear_constraints, 2, True)
        np.testing.assert_array_equal(
            constraints.a_ub,
            [[1.0, 1.0], [-1.0, -1.0]],
        )
        np.testing.assert_array_equal(constraints.b_ub, [1.0, 0.0])
        np.testing.assert_array_equal(constraints.a_eq, [[2.0, 1.0]])
        np.testing.assert_array_equal(constraints.b_eq, [1.0])
        assert constraints.m_ub == 2
        assert constraints.m_eq == 1
        np.testing.assert_allclose(
            constraints.maxcv([0.5, 0.0]),
            0.0,
            atol=1e-15,
        )

    def test_nan(self):
        linear_constraints = [LinearConstraint([[1.0, np.nan]], [0.0], [1.0])]
        constraints = LinearConstraints(linear_constraints, 2, True)
        np.testing.assert_array_equal(
            constraints.a_ub,
            [[1.0, 0.0], [-1.0, 0.0]],
        )
        np.testing.assert_array_equal(constraints.b_ub, [1.0, 0.0])
        assert constraints.m_ub == 2
        assert constraints.m_eq == 0
        linear_constraints = [LinearConstraint([[1.0, 1.0]], [np.nan], [1.0])]
        constraints = LinearConstraints(linear_constraints, 2, True)
        np.testing.assert_array_equal(constraints.a_ub, [[1.0, 1.0]])
        np.testing.assert_array_equal(constraints.b_ub, [1.0])
        assert constraints.m_ub == 1
        assert constraints.m_eq == 0
        linear_constraints = [LinearConstraint([[1.0, 1.0]], [0.0], [np.nan])]
        constraints = LinearConstraints(linear_constraints, 2, True)
        np.testing.assert_array_equal(constraints.a_ub, [[-1.0, -1.0]])
        np.testing.assert_array_equal(constraints.b_ub, [0.0])
        assert constraints.m_ub == 1
        assert constraints.m_eq == 0

    def test_inf(self):
        linear_constraints = [
            LinearConstraint([[1.0, 1.0]], [0.0], [np.inf]),
        ]
        constraints = LinearConstraints(linear_constraints, 2, True)
        np.testing.assert_array_equal(constraints.a_ub, [[-1.0, -1.0]])
        np.testing.assert_array_equal(constraints.b_ub, [0.0])
        assert constraints.m_ub == 1


class TestNonlinearConstraint:

    def test_simple(self):
        nonlinear_constraints = [
            NonlinearConstraint(np.cos, [-0.5, -0.5], [0.0, 0.0]),
            NonlinearConstraint(np.sin, [1.0, 1.0], [1.0, 1.0]),
            NonlinearConstraint(np.tan, -np.inf, np.inf),
            NonlinearConstraint(lambda x: np.inner(x, x) - 1.0, 0, 0),
        ]
        constraints = NonlinearConstraints(nonlinear_constraints, False, True)
        assert constraints.n_eval == 0
        x = [0.5, 0.5]
        c_ub, c_eq = constraints(x)
        np.testing.assert_allclose(
            c_ub,
            np.block([-0.5 - np.cos(x), np.cos(x)]),
            atol=1e-15,
        )
        np.testing.assert_allclose(
            c_eq,
            np.block([np.sin(x) - 1.0, np.inner(x, x) - 1.0]),
            atol=1e-15,
        )
        assert constraints.n_eval == 1
        assert constraints.m_ub == 4
        assert constraints.m_eq == 3
        np.testing.assert_allclose(
            constraints.maxcv(x, c_ub, c_eq),
            max(np.max(np.abs(c_eq)), np.max(c_ub)),
            atol=1e-15,
        )
        np.testing.assert_array_equal(
            constraints.maxcv(x, c_ub, c_eq),
            constraints.maxcv(x),
        )

    def test_args(self):
        nonlinear_constraints = [
            {"fun": lambda x, c: c * np.cos(x),
             "type": "ineq", "args": (2.0,)},
            {"fun": lambda x, c: c * np.sin(x),
             "type": "eq", "args": (2.0,)},
        ]
        x = [0.5, 0.5]
        nonlinear_constraints = standardize_constraints(
            nonlinear_constraints, x, "new"
        )
        constraints = NonlinearConstraints(nonlinear_constraints, False, True)
        c_ub, c_eq = constraints(x)
        np.testing.assert_allclose(
            c_ub,
            np.block([-2.0 * np.cos(x)]),
            atol=1e-15,
        )
        np.testing.assert_allclose(
            c_eq,
            np.block([2.0 * np.sin(x)]),
            atol=1e-15,
        )

    def test_verbose(self, capsys):
        nonlinear_constraints = [
            NonlinearConstraint(np.cos, [-0.5, -0.5], [0.0, 0.0]),
        ]
        constraints = NonlinearConstraints(nonlinear_constraints, True, True)
        x = np.array([1.5, 1.5])
        constraints(x)
        captured = capsys.readouterr()
        with np.printoptions(**PRINT_OPTIONS):
            assert captured.out == f"cos({x}) = {np.cos(x)}\n"

    def test_exceptions(self):
        nonlinear_constraints = [
            NonlinearConstraint(np.cos, [-0.5, -0.5], [0.0, 0.0]),
            NonlinearConstraint(np.sin, [1.0, 1.0], [1.0, 1.0]),
        ]
        constraints = NonlinearConstraints(nonlinear_constraints, False, True)
        with pytest.raises(ValueError):
            constraints.m_ub
        with pytest.raises(ValueError):
            constraints.m_eq


class TestProblem(BaseTest):

    def test_simple(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints(
            [LinearConstraint([[1.0, 1.0]], [0.0], [1.0])],
            2,
            True,
        )
        nonlinear_constraints = NonlinearConstraints(
            [NonlinearConstraint(np.cos, [-0.5, -0.5], [0.0, 0.0])],
            False,
            True,
        )
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        assert problem.n_eval == 0
        x = [0.5, 0.5]
        fun, c_ub, c_eq = problem(x)
        np.testing.assert_allclose(fun, self.rosen(x), atol=1e-15)
        np.testing.assert_allclose(
            c_ub,
            np.block([-0.5 - np.cos(x), np.cos(x)]),
            atol=1e-15,
        )
        assert c_eq.size == 0
        assert problem.n == 2
        assert problem.n_orig == 2
        np.testing.assert_array_equal(problem.x0, [0.0, 0.0])
        assert problem.n_eval == 1
        assert problem.fun_name == "rosen"
        np.testing.assert_array_equal(problem.bounds.xl, [0.0, 0.0])
        np.testing.assert_array_equal(problem.bounds.xu, [1.0, 1.0])
        np.testing.assert_array_equal(
            problem.linear.a_ub,
            [[1.0, 1.0], [-1.0, -1.0]],
        )
        np.testing.assert_array_equal(problem.linear.b_ub, [1.0, 0.0])
        assert problem.linear.a_eq.size == 0
        assert problem.linear.b_eq.size == 0
        assert problem.m_bounds == 4
        assert problem.m_linear_ub == 2
        assert problem.m_linear_eq == 0
        assert problem.m_nonlinear_ub == 4
        assert problem.m_nonlinear_eq == 0
        assert problem.fun_history.size == 0
        assert problem.maxcv_history.size == 0
        assert problem.type == "nonlinearly constrained"
        assert not problem.is_feasibility
        np.testing.assert_allclose(
            problem.maxcv(x),
            max(bounds.maxcv(x), linear_constraints.maxcv(x), np.max(c_ub)),
            atol=1e-15,
        )
        x_best, fun_best, maxcv_best = problem.best_eval(0.0)
        np.testing.assert_array_equal(x_best, x)
        np.testing.assert_allclose(fun_best, self.rosen(x), atol=1e-15)
        np.testing.assert_allclose(maxcv_best, problem.maxcv(x), atol=1e-15)

    def test_scale(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints(
            [LinearConstraint([[1.0, 1.0]], [0.0], [1.0])],
            2,
            True,
        )
        nonlinear_constraints = NonlinearConstraints(
            [NonlinearConstraint(np.cos, [-0.5, -0.5], [0.0, 0.0])],
            False,
            True,
        )
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            True,
            False,
            0,
            1,
            True,
        )
        np.testing.assert_array_equal(problem.bounds.xl, [-1.0, -1.0])
        np.testing.assert_array_equal(problem.bounds.xu, [1.0, 1.0])
        np.testing.assert_allclose(problem.x0, [-1.0, -1.0], atol=1e-15)
        np.testing.assert_allclose(
            problem.linear.a_ub,
            [[0.5, 0.5], [-0.5, -0.5]],
            atol=1e-15,
        )
        np.testing.assert_allclose(problem.linear.b_ub, [0.0, 1.0], atol=1e-15)
        x = np.array([0.5, 0.5])
        fun, c_ub, _ = problem(2.0 * x - 1.0)
        np.testing.assert_allclose(fun, self.rosen(x), atol=1e-15)
        np.testing.assert_allclose(
            c_ub,
            np.block([-0.5 - np.cos(x), np.cos(x)]),
            atol=1e-15,
        )

    def test_barrier(self):
        obj = ObjectiveFunction(lambda x: np.nan, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints([], 2, True)
        nonlinear_constraints = NonlinearConstraints(
            [
                NonlinearConstraint(lambda x: np.nan, [0.0], [1.0]),
                NonlinearConstraint(lambda x: np.nan, [0.0], [0.0]),
            ],
            False,
            True,
        )
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        fun, c_ub, c_eq = problem([0.5, 0.5])
        assert np.isfinite(fun)
        assert np.all(np.isfinite(c_ub))
        assert np.all(np.isfinite(c_eq))

    def test_history(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints([], 2, True)
        nonlinear_constraints = NonlinearConstraints(
            [NonlinearConstraint(np.cos, [-0.5, -0.5], [0.0, 0.0])],
            False,
            True,
        )
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            True,
            2,
            1,
            True,
        )
        x = [0.5, 0.5]
        problem(x)
        np.testing.assert_array_equal(problem.fun_history, [self.rosen(x)])
        np.testing.assert_allclose(
            problem.maxcv_history,
            [nonlinear_constraints.maxcv(x)],
            atol=1e-15,
        )
        problem(x)
        np.testing.assert_array_equal(
            problem.fun_history,
            2 * [self.rosen(x)],
        )
        np.testing.assert_allclose(
            problem.maxcv_history,
            2 * [nonlinear_constraints.maxcv(x)],
            atol=1e-15,
        )
        problem(x)
        np.testing.assert_array_equal(
            problem.fun_history,
            2 * [self.rosen(x)],
        )
        np.testing.assert_allclose(
            problem.maxcv_history,
            2 * [nonlinear_constraints.maxcv(x)],
            atol=1e-15,
        )

    def test_filter(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints(
            [LinearConstraint([[1.0, 1.0]], [1.0], [1.0])],
            2,
            True,
        )
        nonlinear_constraints = NonlinearConstraints([], False, True)
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        problem([1.0, 1.0])
        x, _, _ = problem.best_eval(1e3)
        np.testing.assert_allclose(x, [1.0, 1.0], atol=1e-15)
        problem([0.25, 0.75])
        x, _, _ = problem.best_eval(1e3)
        np.testing.assert_allclose(x, [0.25, 0.75], atol=1e-15)
        problem([0.5, 0.5])
        x, _, _ = problem.best_eval(1e3)
        np.testing.assert_allclose(x, [0.5, 0.5], atol=1e-15)

    def test_callback(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints([], 2, True)
        nonlinear_constraints = NonlinearConstraints([], False, True)

        def callback(intermediate_result):
            if intermediate_result.fun < 1e-3:
                raise StopIteration

        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            callback,
            0.0,
            False,
            False,
            0,
            2,
            True,
        )
        problem([0.0, 0.0])
        with pytest.raises(CallbackSuccess):
            problem([1.0, 1.0])

        def callback(xk):
            if np.all(xk > 0.5):
                raise CallbackSuccess

        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            callback,
            0.0,
            False,
            False,
            0,
            2,
            True,
        )
        problem([0.0, 0.0])
        with pytest.raises(CallbackSuccess):
            problem([1.0, 1.0])

    def test_type(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds(2 * [-np.inf], 2 * [np.inf]))
        linear_constraints = LinearConstraints([], 2, True)
        nonlinear_constraints = NonlinearConstraints([], False, True)
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        assert problem.type == "nonlinearly constrained"
        problem(problem.x0)
        assert problem.type == "unconstrained"
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        problem(problem.x0)
        assert problem.type == "bound-constrained"
        linear_constraints = LinearConstraints(
            [LinearConstraint([[1.0, 1.0]], [0.0], [1.0])],
            2,
            True,
        )
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        problem(problem.x0)
        assert problem.type == "linearly constrained"

    def test_feasibility_problem(self):
        obj = ObjectiveFunction(None, False, True)
        bounds = BoundConstraints(Bounds(2 * [-np.inf], 2 * [np.inf]))
        linear_constraints = LinearConstraints([], 2, True)
        nonlinear_constraints = NonlinearConstraints([], False, True)
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        assert problem.is_feasibility
        obj = ObjectiveFunction(self.rosen, False, True)
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            1,
            True,
        )
        assert not problem.is_feasibility

    def test_best_eval(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints(
            [LinearConstraint([[1.0, 1.0]], [0.0], [1.0])],
            2,
            True,
        )
        nonlinear_constraints = NonlinearConstraints(
            [NonlinearConstraint(np.cos, [-0.5, -0.5], [0.0, 0.0])],
            False,
            True,
        )
        problem = Problem(
            obj,
            [0.0, 0.0],
            bounds,
            linear_constraints,
            nonlinear_constraints,
            None,
            0.0,
            False,
            False,
            0,
            2,
            True,
        )
        x, _, _ = problem.best_eval(1e3)
        np.testing.assert_array_equal(x, [0.0, 0.0])

    def test_exceptions(self):
        obj = ObjectiveFunction(self.rosen, False, True)
        bounds = BoundConstraints(Bounds([0.0, 0.0], [1.0, 1.0]))
        linear_constraints = LinearConstraints([], 2, True)
        nonlinear_constraints = NonlinearConstraints([], False, True)
        with pytest.raises(TypeError):
            Problem(
                obj,
                [0.0, 0.0],
                bounds,
                linear_constraints,
                nonlinear_constraints,
                1.0,
                0.0,
                False,
                False,
                0,
                1,
                True,
            )
        wrong_bounds = BoundConstraints(Bounds([0.0], [1.0]))
        with pytest.raises(ValueError):
            Problem(
                obj,
                [0.0, 0.0],
                wrong_bounds,
                linear_constraints,
                nonlinear_constraints,
                None,
                0.0,
                False,
                False,
                0,
                1,
                True,
            )
        wrong_linear_constraints = LinearConstraints(
            [LinearConstraint(1.0, 1.0, 1.0)],
            1,
            True,
        )
        with pytest.raises(ValueError):
            Problem(
                obj,
                [0.0, 0.0],
                bounds,
                wrong_linear_constraints,
                nonlinear_constraints,
                None,
                0.0,
                False,
                False,
                0,
                1,
                True,
            )
